import numpy as np
from scipy.stats import rankdata

def transfer_to_ranking_score(gw):
    '''
    transfer the gene weight to ranking
    '''
    # gw = adata.layers['geneWeight'].toarray()
    od = gw.shape[1]-rankdata(gw,axis=1)+1
    score = 1+1/np.power(od,0.5)
    
    return score


import pickle
import pandas as pd
import matplotlib as mpl
import numpy as np
import matplotlib.pyplot as plt
import scanpy as sc

def cell_type_composition(adata, cell_type_key, stage_key,ax=None,dpi=300,show_cutoff = 0.04, colormaps='Spectral',category_colors=None, save=None):
    '''
    Plot the cell type composition of each stage

    Parameters
    ----------
    adata : AnnData object
        Annotated data matrix.
    cell_type_key : str
        Key for cell type column in adata.obs.
    stage_key : str
        Key for stage column in adata.obs.
    ax : matplotlib axis, optional
        The default is None.
    dpi : int, optional
        The default is 300.
    show_cutoff : float, optional
        The default is 0.04.
    colormaps : str, optional
        The default is 'Spectral'.
    category_colors : list, optional
        The default is None.
    save : str, optional
        Path to save the figure. The default is None.
    '''

    all_types = adata.obs[cell_type_key].unique().tolist()
    all_types = sorted(all_types)
    stage_keys = adata.obs[stage_key].unique().tolist()
    stage_keys = sorted(stage_keys)
    stage_keys = stage_keys[::-1]
    stage_types = {key :[] for key in stage_keys}
    total_stage = len(stage_keys)
    for i in stage_keys:
        stage_adata = adata[adata.obs[stage_key] == i]
        stage_cells = len(stage_adata)
        for each in all_types:
            stage_types[i].append(len(stage_adata[stage_adata.obs[cell_type_key] == each])/stage_cells)
    multiplier = 0
    width = 0.3
    if ax is None:
        fig,ax = plt.subplots(figsize=(15, 3),dpi=dpi)

    #transform the type of list to array
    # y_pos = [0, 1]
    for i in stage_types.keys():
        stage_types[i] = np.array(stage_types[i])
    bottom = 0
    data = np.array(list(stage_types.values()))
    data_cum = data.cumsum(axis=1)
    if category_colors is None:
        category_colors = mpl.colormaps[colormaps](np.linspace(0.1, 1, data.shape[1]))
    for j, (types,c) in enumerate(zip(all_types,category_colors)):
        widths = data[:,j]
        starts = data_cum[:, j] - widths 
        bc = ax.barh(list(stage_types.keys()), widths, left = starts, label=types,color=c,height=0.8)
        labels = ["" if v > show_cutoff else "" for v in widths]  
        ax.bar_label(bc, labels=labels,label_type='center',fmt='%.2f',fontsize=9,color='black')
    ax.legend(ncol=len(all_types)//2+1, bbox_to_anchor=(0.1, 1.3),
                loc='upper left', fontsize='x-small')
    # threshold = 0.15
    for c in ax.containers:#comments out to disable percentage on the plot
        #v format %.2f
        labels = [f'%.2f%%'%(v*100) if v > show_cutoff else "" for v in c.datavalues] #comments out to disable percentage on the plot
        # labels = [v if v > 0.05 else "" for v in c.datavalues]  
        ax.bar_label(c, labels=labels, label_type="center",fmt='%.2f')#comments out to disable percentage on the plot
        # ax.bar_label(bc, labels=weight_count,label_type='center',fontsize=9)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.get_xaxis().set_ticks([])
    ax.grid(False) 
    if save is not None:
        plt.savefig(save)
    else:
        plt.show()


import numpy as np
import scanpy as sc
import gc
import os
from .distDistance import *
def nodesDistance(rep1,rep2,topgene1,topgene2):
    '''
    calculate the distance between two stages
    
    parameters
    -------------------
    rep1: list
        The representation of clusters in stage 1
    rep2: list
        The representation of clusters in stage 2
    topgene1: list
        Top 100 differential gene of clusters in stage 1
    topgene2: list
        Top 100 differential gene of clusters in stage 2
    
    return
    -------------------
    distance: list
        A list of normalized distance of clusters between two stages
    '''
    distance = [[] for _ in range(len(rep2))]
    for i in range(len(rep2)):
        for j in range(len(rep1)):
            gaussiankl_1 = calculateKL(rep2[i],rep1[j])
            gaussiankl_2 = calculateKL(rep1[j],rep2[i])
            gaussiankl = (gaussiankl_1 + gaussiankl_2)/2
            similarityDE = getSimilarity(topgene2,topgene1,i,j)
            distance[i].append([gaussiankl , similarityDE])
    for i in range(len(distance)):
        distance[i] = normalizeDistance(distance[i])
    return distance
def connectNodes(distances,cutoff = 0.05):
    '''
    Connect the clusters in two stages with smallest distance and p-value < cut-off

    parameters
    -------------------
    distances: list
        The list of distance between two stages
    cutoff: float
        The cutoff of p-value

    return
    -------------------
    edges: list
        The edges between two stages
    '''
    edges = []
    for i in range(len(distances)):
        leftend = np.argmin(distances[i])
        # temp = distances[i].copy()
        pval = norm.cdf(distances[i][leftend],loc = np.mean(distances),scale = np.std(distances))
        #p*count/(count-idx)
        # q_val = pval * len(temp)/
        
        if pval < cutoff: #if pval < 0.01 can connect the two clusters across two stages
            edges.append([leftend,i])
    return edges

def buildEdges(stage1,stage2,cutoff = 0.05):
    '''
    calculate the distance between two stages and connect the clusters in two stages with smallest distance
    
    parameters
    -------------------

    stage1: anndata
        The data of the first selected stage
    stage2: anndata
        The data of the second selected stage
    cutoff: float
        The cutoff of p-value
    
    return
    -------------------
    edges: list
        The edges between two stages
    '''
    adata1 = sc.read_h5ad('./stagedata/%d.h5ad'%stage1)
    adata2 = sc.read_h5ad('./stagedata/%d.h5ad'%stage2)
    reps = np.load('./stagedata/rep.npy',allow_pickle=True)
    
    rep1 = reps[stage1]
    rep2 = reps[stage2]
    topgene1 = adata1.uns['topGene']
    topgene2 = adata2.uns['topGene']
    distance = nodesDistance(rep1,rep2,topgene1,topgene2)
    edges = connectNodes(distance,cutoff)
    return edges
def buildEdges(stage1,stage2,midpath,iteration,cutoff = 0.05):
    '''
    calculate the distance between two stages and connect the clusters in two stages with smallest distance with midpath in iterative training
    
    parameters
    ------------------- 

    midpath: str
        The task name
    iteration: int
        The iteration of the training
    stage1: anndata
        The data of the first selected stage
    stage2: anndata
        The data of the second selected stage
    cutoff: float
        The cutoff of p-value

    return
    -------------------
    edges: list
        The edges between two stages
    '''
    # print(stage1,stage2)
    adata1 = sc.read_h5ad(os.path.join(midpath,str(iteration)+'/stagedata/%d.h5ad'%stage1))
    adata2 = sc.read_h5ad(os.path.join(midpath,str(iteration)+'/stagedata/%d.h5ad'%stage2))
    reps = np.load(os.path.join(midpath,str(iteration)+'/stagedata/rep.npy'),allow_pickle=True)
    
    rep1 = reps[stage1]
    rep2 = reps[stage2]
    topgene1 = adata1.uns['topGene']

    topgene2 = adata2.uns['topGene']
    distance = nodesDistance(rep1,rep2,topgene1,topgene2)
    edges = connectNodes(distance,cutoff)
    return edges
def getandUpadateEdges(total_stage,midpath,iteration):
    '''
    get edges in iterative training.
    
    parameters
    -------------------
    total_stage: int
        The total number of stages
    midpath: str
        The task name
    iteration: int
        The iteration of the training

    return
    -------------------
    edges: list
        The edges between two stages
    '''
    edges = []
    for i in range(total_stage-1):
        edges.append(buildEdges(i,i+1,midpath,iteration))
    updateEdges(edges,midpath,iteration)
    return edges

def updateEdges(edges,midpath,iteration):
    '''
    updata edges to the anndata database, calculate edges changes.

    parameters
    -------------------
    edges: list
        The edges between two stages
    midpath: str
        The task name
    iteration: int
        The iteration of the training

    return
    -------------------
    edges: list
        The edges between two stages

    '''
    newEdges = {}
    for i in range(len(edges)):
        newEdges[str(i)] = edges[i]
    #adata.uns['edges'] = newEdges
    f = open(os.path.join(midpath,str(iteration)+'/edges.txt'),'w')
    f.write(str(newEdges))
    f.close()

def reupdateAttributes(adata, stage, results):
    '''
    update gaussian and gamma rep, top 100 differential genes, cell types of clusters to anndata

    parameters
    -------------------
    adata: anndata 
        The single-cell data
    stage: int
        The selected stage

    results: list
        A list contained top differential genes and cell types of clusters
    
    return
    -------------------
    adata: anndata
        updated anndata of input single-cell data
    '''
    stageids = adata.obs[adata.obs['stage'] == stage].index.tolist()
    tempadata = adata[stageids]
    adata.uns['rep'][str(stage)] = results[0]
    adata.uns['topGene'][str(stage)] = results[1]
    adata.uns['clusterType'][str(stage)] = results[2]
    return adata



'''
The perturbations module allows to perform the pathway and compounds in-silico perturbations and obatin the perturbation results.
'''

from .compounds import get_top_compounds
from .pathways import get_top_pathways


# get top markers for each track from IDREM results, the requirements for top markers are:
# 1. the tendency of the marker is monotonic
# 2. the log2FC of the marker is larger than a cutoff
import json
import os
import numpy as np
import pandas as pd
from scipy.stats import norm
from ..dynamic_regulatory_networks.processTFs import readIdremJson
def scoreAgainstBackground(background,input,all=False,mean=None,std=None):
    '''
    Calculate the p-value of the input gene expression change based on the background gene expression change.
    
    parameters
    --------------------
    background: the background gene expression change
    input: the input gene expression change
    all: if all is True, the input gene expression change will be compared with all the background gene expression changes. Otherwise, the input gene expression change will be compared with each background gene expression change.
    mean: the mean of the background gene expression change
    std: the standard deviation of the background gene expression change

    return
    --------------------
    cdf: np.float
        p-value of the input gene expression change
    '''

    if all:
        background = background.reshape(-1,1)

    # print(background.mean(axis=0), background.std(axis=0))
    if mean is None:
        cdf = norm.cdf(input, loc=background.mean(axis=0), scale=background.std(axis=0))
    else:
        cdf = norm.cdf(input, loc=mean, scale=std)
    
    return cdf

def getTopMarkers(idrem, background,filename, cutoff=None,topN=None,one_dist=False):
    '''
    **not updated**
    '''
    topMarkers = {}
    tt = readIdremJson(idrem,filename)

    background = np.array(background)
    stages = len(filename.split('.')[0].split('-'))
    
    temp = np.array(tt[8])
    idrem_genes = np.array(temp[1:,0].tolist())
    tendency = temp[1:,1].astype(float)
    for i in range(2,stages):
        tendency*=temp[1:,i].astype(float)
    tendency[tendency <0] = 0
    index = [i for i, x in enumerate(tendency) if x <= 0]
    change = temp[1:,stages].astype(float) - temp[1:,1].astype(float)
    stage_values = []
    for i in range(0,stages):
        stage_values.append(temp[1:,i+1].astype(float))

    stage0 = temp[1:,1].astype(float)
    stage1 = temp[1:,2].astype(float)
    stage2 = temp[1:,3].astype(float)
    stage3 = temp[1:,4].astype(float)
    
    change[index] = 0
    if cutoff is not None:
        topMarkers['increasing'] = {}
        topMarkers['decreasing'] = {}
        increasing_stop = np.where(change > 0)[0]
        decreasing_stop = np.where(change < 0)[0]
        temp_change = change[increasing_stop]
        temp_names = idrem_genes[increasing_stop]
        temp_stage0 = stage0[increasing_stop]
        temp_stage1 = stage1[increasing_stop]
        temp_stage2 = stage2[increasing_stop]
        temp_stage3 = stage3[increasing_stop]
        temp_background = background[:,increasing_stop]

        
        increasing_stop = temp_change.argsort()[::-1]
        # increasing_pval = scoreAgainstBackground(background[:,increasing_stop],temp_change)
        
        topMarkers['increasing']['gene'] = {}
        topMarkers['increasing']['log2fc'] = {}
        topMarkers['increasing']['rank'] = {}
        topMarkers['increasing']['stage0'] = {}
        topMarkers['increasing']['stage1'] = {}
        topMarkers['increasing']['stage2'] = {}
        topMarkers['increasing']['stage3'] = {}
        topMarkers['increasing']['qval'] = {}
        count = 0
        pvals = []
        in_pvals = []
        if one_dist:
            temp_background = temp_background.reshape(-1,1)
            mean = temp_background.mean()
            std = temp_background.std()
        for i, each in enumerate(increasing_stop):
            if one_dist:
                increasing_pval = scoreAgainstBackground(temp_background,temp_change[each],mean = mean, std = std)
            else:
                increasing_pval = scoreAgainstBackground(temp_background[:, each],temp_change[each])
                if increasing_pval < (1-cutoff):
                    continue
                increasing_pval = scoreAgainstBackground(temp_background,temp_change[each],all=True)
            if increasing_pval < (1-cutoff):
                continue
            in_pvals.append(each)
            pvals.append(1-increasing_pval)
            # topMarkers['increasing']['gene'][str(count)] = temp_names[each]
            # topMarkers['increasing']['stage0'][str(count)] = temp_stage0[each]
            # topMarkers['increasing']['stage1'][str(count)] = temp_stage1[each]
            # topMarkers['increasing']['stage2'][str(count)] = temp_stage2[each]
            # topMarkers['increasing']['stage3'][str(count)] = temp_stage3[each]
            # topMarkers['increasing']['log2fc'][str(count)] = temp_change[each]
            # topMarkers['increasing']['rank'][str(count)] = count+1
            # topMarkers['increasing']['pval'][str(count)] = 1-increasing_pval#[each]
            count+=1
        count_1 = 0
        #sort pvals by sorted
        pvals = np.array(pvals).reshape(-1)

        in_pvals = np.array(in_pvals).reshape(-1)
        pvals = pvals[pvals.argsort()]

        in_pvals = in_pvals[pvals.argsort()]
        for i in range(count):
            
            q_val = pvals[i] * count / (i+1)
            if q_val > cutoff:
                continue
            each = in_pvals[i]
            topMarkers['increasing']['gene'][str(count_1)] = temp_names[each]
            topMarkers['increasing']['stage0'][str(count_1)] = temp_stage0[each]
            topMarkers['increasing']['stage1'][str(count_1)] = temp_stage1[each]
            topMarkers['increasing']['stage2'][str(count_1)] = temp_stage2[each]
            topMarkers['increasing']['stage3'][str(count_1)] = temp_stage3[each]
            topMarkers['increasing']['log2fc'][str(count_1)] = temp_change[each]
            topMarkers['increasing']['rank'][str(count_1)] = count_1+1
            topMarkers['increasing']['qval'][str(count_1)] = q_val
            count_1+=1
        temp_change = change[decreasing_stop]
        temp_names = idrem_genes[decreasing_stop]
        temp_stage0 = stage0[decreasing_stop]
        temp_stage1 = stage1[decreasing_stop]
        temp_stage2 = stage2[decreasing_stop]
        temp_stage3 = stage3[decreasing_stop]
        temp_background = background[:,decreasing_stop]
        decreasing_stop = temp_change.argsort()
        # decreasing_pval = scoreAgainstBackground(background[:,decreasing_stop],temp_change)
        
        topMarkers['decreasing']['gene'] = {}
        topMarkers['decreasing']['log2fc'] = {}
        topMarkers['decreasing']['rank'] = {}
        topMarkers['decreasing']['stage0'] = {}
        topMarkers['decreasing']['stage1'] = {}
        topMarkers['decreasing']['stage2'] = {}
        topMarkers['decreasing']['stage3'] = {}
        topMarkers['decreasing']['qval'] = {}
        count = 0
        pvals = []
        in_pvals = []
        if one_dist:
            temp_background = temp_background.reshape(-1,1)
            mean = temp_background.mean()
            std = temp_background.std()
        for i, each in enumerate(decreasing_stop):
            if one_dist:
                decreasing_pval = scoreAgainstBackground(temp_background,temp_change[each],mean = mean, std = std)
            else:
                decreasing_pval = scoreAgainstBackground(temp_background[:, each],temp_change[each])
                if decreasing_pval > cutoff:
                    continue
                decreasing_pval = scoreAgainstBackground(temp_background,temp_change[each],all=True)
            if decreasing_pval > cutoff:
                continue
            in_pvals.append(each)
            pvals.append(decreasing_pval)
            # topMarkers['decreasing']['gene'][str(count)] = temp_names[each]
            # topMarkers['decreasing']['log2fc'][str(count)] = temp_change[each]
            # topMarkers['decreasing']['rank'][str(count)] = count+1 
            # topMarkers['decreasing']['stage0'][str(count)] = temp_stage0[each]
            # topMarkers['decreasing']['stage1'][str(count)] = temp_stage1[each]
            # topMarkers['decreasing']['stage2'][str(count)] = temp_stage2[each]
            # topMarkers['decreasing']['stage3'][str(count)] = temp_stage3[each]
            # topMarkers['decreasing']['pval'][str(count)] = decreasing_pval#[each]
            count+=1
        count_1 = 0
        
        pvals = np.array(pvals).reshape(-1)
        in_pvals = np.array(in_pvals).reshape(-1)
        pvals = pvals[pvals.argsort()]
        in_pvals = in_pvals[pvals.argsort()]

        for i in range(count):
            q_val = pvals[i] * count/(i+1)  #1 - (1 - topMarkers['decreasing']['pval'][str(i)]) * (count/(i+1))
            if q_val > cutoff:
                continue
            each = in_pvals[i]
            topMarkers['decreasing']['gene'][str(count_1)] = temp_names[each]
            topMarkers['decreasing']['log2fc'][str(count_1)] = temp_change[each]
            topMarkers['decreasing']['rank'][str(count_1)] = count_1+1 
            topMarkers['decreasing']['stage0'][str(count_1)] = temp_stage0[each]
            topMarkers['decreasing']['stage1'][str(count_1)] = temp_stage1[each]
            topMarkers['decreasing']['stage2'][str(count_1)] = temp_stage2[each]
            topMarkers['decreasing']['stage3'][str(count_1)] = temp_stage3[each]
            topMarkers['decreasing']['qval'][str(count_1)] = q_val
            count_1+=1
    else: # if cutoff is not given, return ranked list
        topMarkers['increasing'] = {}
        topMarkers['decreasing'] = {}
        increasing_stop = np.where(change > 0)[0]
        decreasing_stop = np.where(change < 0)[0]
        temp_change = change[increasing_stop]
        temp_names = idrem_genes[increasing_stop]
        temp_stage0 = stage0[increasing_stop]
        temp_stage1 = stage1[increasing_stop]
        temp_stage2 = stage2[increasing_stop]
        temp_stage3 = stage3[increasing_stop]
        increasing_stop = temp_change.argsort()[::-1]
        topMarkers['increasing']['gene'] = {}
        topMarkers['increasing']['log2fc'] = {}
        topMarkers['increasing']['rank'] = {}
        topMarkers['increasing']['stage0'] = {}
        topMarkers['increasing']['stage1'] = {}
        topMarkers['increasing']['stage2'] = {}
        topMarkers['increasing']['stage3'] = {}
        if topN is not None:
            for i, each in enumerate(increasing_stop[:topN]):
                topMarkers['increasing']['gene'][str(i)] = temp_names[each]
                topMarkers['increasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['increasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['increasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['increasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['increasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['increasing']['rank'][str(i)] = i+1
        else:
            for i, each in enumerate(increasing_stop):
                topMarkers['increasing']['gene'][str(i)] = temp_names[each]
                topMarkers['increasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['increasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['increasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['increasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['increasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['increasing']['rank'][str(i)] = i+1
        temp_change = change[decreasing_stop]
        temp_names = idrem_genes[decreasing_stop]
        temp_stage0 = stage0[decreasing_stop]
        temp_stage1 = stage1[decreasing_stop]
        temp_stage2 = stage2[decreasing_stop]
        temp_stage3 = stage3[decreasing_stop]
        decreasing_stop = temp_change.argsort()
        topMarkers['decreasing']['gene'] = {}
        topMarkers['decreasing']['log2fc'] = {}
        topMarkers['decreasing']['rank'] = {}
        topMarkers['decreasing']['stage0'] = {}
        topMarkers['decreasing']['stage1'] = {}
        topMarkers['decreasing']['stage2'] = {}
        topMarkers['decreasing']['stage3'] = {}
        if topN is not None:
            for i, each in enumerate(decreasing_stop[:topN]):
                topMarkers['decreasing']['gene'][str(i)] = temp_names[each]
                topMarkers['decreasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['decreasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['decreasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['decreasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['decreasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['decreasing']['rank'][str(i)] = i+1

        else:
            for i, each in enumerate(decreasing_stop):
                topMarkers['decreasing']['gene'][str(i)] = temp_names[each]
                topMarkers['decreasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['decreasing']['rank'][str(i)] = i+1 
                topMarkers['decreasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['decreasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['decreasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['decreasing']['stage3'][str(i)] = temp_stage3[each]


    return topMarkers
def getTopMarkersFromIDREM(path, background,cutoff=None,topN=None,one_dist=False):
    '''
    Get the top markers for each track from IDREM results.
    parameters
    --------------------
    path: str
        the directory to the IDREM results.
    background: np.array
        the background gene expression change
    cutoff: float
        the cutoff for p-value. Default is None.
    topN: int
        the number of top markers to return. Default is None.
    one_dist: bool
        whether to consider all the background gene expression changes as one distribution. Default is False.

    return
    ----------------
    out:    dict
        a dictionary of top markers for each track.
    '''
    out = {}
    filenames = os.listdir(path)
    for each in filenames:
        name = each
        if each[0] != '.':
            each = each.split('.')[0]#.split('-')[-1].split('n')
            if one_dist:
                out[each] = getTopMarkers(path,background,name,cutoff,topN,one_dist=one_dist)
            else:
                out[each] = getTopMarkers(path,background[each],name,cutoff,topN)
    return out

def runGetProgressionMarkercsv(directory,background, save_dir, topN=None,cutoff=None):
    '''
    Get the top markers for each track from IDREM results and save as a csv file.
    
    parameters
    --------------------
    directory: str
        the directory to the IDREM results.
    background: str
        the directory to the background gene expression change.
    cutoff: float
        the cutoff for p-value. Default is None.
    topN: int
        the number of top markers to return. Default is None.
    save_dir: str
        the directory to save the csv file.

    return
    ---------------
    None
    '''
    background = dict(np.load(background,allow_pickle=True).tolist())
    out = getTopMarkersFromIDREM(directory,background,topN=topN,cutoff=cutoff)
    results = pd.json_normalize(out)
    results.columns = results.columns.str.split(".").map(tuple)
    results = results.stack([0,1,2]).reset_index(0, drop=True)
    results = results.transpose()
    results = results.reset_index()
    results['index'] = results['index'].astype('int32')
    results = results.set_index(['index']).sort_index()
    # results = results.transpose()

    results = results.sort_index()
    # print(results)
    # return results
    results.to_csv(os.path.join(save_dir,'mesProgressionMarker_pval_twofilters.csv')) 


def runGetProgressionMarker(directory,background, cutoff=0.05, topN=None):
    '''
    Get the top markers for each track from IDREM results.
    
    parameters
    --------------------
    directory: str
        the directory to the IDREM results.
    background: str
        the directory to the background gene expression change.
    cutoff: float
        the cutoff for p-value. Default is 0.05.
    topN: int
        the number of top markers to return. Default is None.

    return
    ---------------
    out:    dict
        a dictionary of top markers for each track.
    '''
    out = getTopMarkersFromIDREM(directory,background,cutoff = cutoff,topN=topN)
    return out


def runGetProgressionMarker_one_dist(directory,background,size, cutoff=0.05, topN=None):
    '''
    Get the top markers for each track from IDREM results and consider the whole background as one distribution.
    
    parameters
    --------------------
    directory: str
        the directory to the IDREM results.
    background: str
        the directory to the background gene expression change.
    cutoff: float
        the cutoff for p-value. Default is 0.05.
    topN: int
        the number of top markers to return. Default is None.

    return
    ---------------
    out:    
        a dictionary of top markers for each track.
    '''

    one_dist  = []
    for each in background.keys(): 
        one_dist.append(np.array(background[each]))
    background = np.array(one_dist).reshape(-1, size)
    out = getTopMarkersFromIDREM(directory,background,cutoff = cutoff,one_dist=True)
    return out


import numpy as np
from sklearn.neighbors import kneighbors_graph 
import scanpy as sc
import torch
import os
def find_neighbourhood(adj, start, end):
    '''
    find and return the neighbourhoods of cells in the batch)
    '''
    # adj = adj.to_dense()
    adj = adj.numpy()

    targets = [i for i in range(start, end)]
    out = targets.copy()
    for each in targets:
        
        out+=list(np.nonzero(adj[each])[0])
    out = list(set(out))
    return out
def new_find_neighbourhood(adj):
    '''
    find and return the neighbourhoods of cells in the batch)
    '''
    # adj = adj.to_dense()
    # adj = adj.numpy()
    out = []
    for each in adj:
        temp = each.indices.tolist()
        out.append(temp)
    return out


def setup_graph(adj):
    '''
    transfer adj into coo and set up self weight
    '''
    # adj.setdiag(1)
    # adj = adj.asformat('coo')
    adj.data = adj.data
#     adj.setdiag(1)
    adj_values = adj.data
    adj_indices = np.vstack((adj.row, adj.col))
    adj_shape = adj.shape
    adj_i = torch.LongTensor(adj_indices)
    adj_v = torch.FloatTensor(adj_values)
    adj = torch.sparse.FloatTensor(adj_i, adj_v, torch.Size(adj_shape))
    return adj
def get_gcn_exp(source_directory,total_stage,neighbors,threads= 20):
    '''
    get the gcn connectivities for each cell
    save stage adata with gcn connectivities in the same directory
    '''

    for i in range(total_stage):
        print('Calculating cell graph for stage %d.....'%i)
        read_path = source_directory+'/%d.h5ad'%i
        temp = sc.read_h5ad(read_path)
        
        if 'gcn_connectivities' in temp.obsp.keys():
            continue

        sc.pp.pca(temp)
        adj = kneighbors_graph(temp.obsm['X_pca'],  neighbors-1, mode='connectivity', include_self=False,n_jobs=threads)
        adj.setdiag(neighbors)#for pcls
        temp.obsp['gcn_connectivities'] = adj
        write_path = os.path.join(source_directory,'%d.h5ad'%i)
        temp.write(write_path, compression='gzip', compression_opts=9)

def get_neighbours(batch_size, adj, cell_loader):
    '''
    return a list of neighbors to speed up training
    '''
    neighbourhoods = []
    for i, x in enumerate(cell_loader):
        # temp_x = placeholder.clone()
        start = i*batch_size
        if (1+i)*batch_size > len(adj):
            end =  len(adj)
        else:
            end = (1+i)*batch_size
        neighbourhoods.append(find_neighbourhood(adj.to_dense(), start,end))
    return neighbourhoods


#this is push back, last cell is push forward
import numpy as np
import os
import random
import gc
import torch
from scipy.sparse import issparse
#import DataLoader from torch
from torch.utils.data import DataLoader
import threading
from ..model.models import VAE
from .analysis_perturbation import perturbationAnalysis
class perturbation:
    def __init__(self, target_directory,model_name,idrem_dir):
        self.model_name = model_name
        self.target_directory = target_directory
        
        self.idrem_dir = idrem_dir
        

        self.adata = self.read_mergeadata()
        self.total_stage = len(set(self.adata.obs['stage']))
        self.tracks = self.getTrackReadOrder()
        self.stageadata = self.read_stagedata()
        
        
        self.hiddenReps = []
        self.perturb_stage_data_mean = []

        
    def read_mergeadata(self):
        read_path = self.target_directory#os.path.join(self.target_directory,'dataset.h5ad')
        mergeadata = read_path#sc.read_h5ad(read_path)
        mergeadata.obs['leiden'] = mergeadata.obs['leiden'].astype('string')
        return mergeadata
    def read_stagedata(self):
        stageadata = []
        self.stage_cluster = {}
        stage_have_clusters = {}
        stage_have_clusters[str(0)] = []
        for i in self.tracks.keys():
            track = self.getTrack(len(self.adata.obs['stage'].unique())-1,i)
            track_name = str(track[0][0])
            stage_have_clusters[str(0)].append(str(track[0][0]))
            for j in range(1,len(track)):
                if str(j) not in stage_have_clusters.keys():
                    stage_have_clusters[str(j)] = []
                stage_have_clusters[str(j)].append(str(track[j][0]))
                track_name += '-' + str(track[j][0])
        self.adata.obs['stage'] = self.adata.obs['stage'].astype('string')
        for i in range(0,4):

            stagedataids = self.adata.obs[self.adata.obs['stage']==str(i)].index.values
            

            adata = self.adata[stagedataids]
            adata.obs['leiden'] = adata.obs['leiden'].astype('string')
            grouped = adata.obs.groupby('leiden')
            self.stage_cluster[str(i)] = {}#{name: adata[group.index.tolist()] for name, group in grouped}
            for name, group in grouped:
                if str(name) not in stage_have_clusters[str(i)]:
                    continue
                else:
                    self.stage_cluster[str(i)][str(name)] = adata[group.index.tolist()]
            stageadata.append(adata)
        return stageadata
    def get_gene_names(self):
        return np.array(list(self.stageadata[0].var.index.values))
    # def save_to_disk(self):
    #         for i, each in enumerate(self.stageadata):
    #             save_path = os.path.join(self.target_directory,'stagedata/concat_%d.h5ad'%(i))
    #             each.write_h5ad(save_path)
    def getDescendants(tempcluster,stage,edges):
        out = []
        for each in tempcluster:
            
            for item in edges[str(stage-1)]:
            
                if each == item[0]:
                    
                    out.append(item[1])
        return out
    def getDistance(self,rep, cluster):

        cluster = cluster.reshape(1,-1)
        cluster = cluster.repeat(rep.shape[0],axis=0)
        return np.linalg.norm(rep-cluster,axis=1)
    def matchSingleClusterGeneDict(self,goadata,gotop):
        gene_dict={}
        for i,each in enumerate(goadata.var.index.tolist()):
            gene_dict[each]=i
        results=[]
        for each in gotop:
            if each not in goadata.var.index.tolist():

                continue
            results.append(gene_dict[each])
        return results
    def getTrack(self,stage, clusterid):
        path = self.idrem_dir#os.path.join(self.target_directory,'idremVizCluster')
        filenames = os.listdir(path) #defalut path

        tempTrack = [[] for _ in range(self.total_stage)]
        for each in filenames:
            temp = each.split('.')[0].split('-')
            for i,item in enumerate(temp):
                temp1 = item.split('n')
                tempTrack[i].append(temp1)
        track = [[] for _ in range(self.total_stage)]
        # open_path = os.path.join(self.target_directory,'edges.txt')
        # edges = eval(open(open_path, 'r').read())
        edges = self.adata.uns['edges']
        for i, each in enumerate(tempTrack[int(stage)]):
            if str(clusterid) in each:
                track[0] = [int(tempTrack[0][i][0])]
                
                tempcluster = clusterid
                for k in range(int(stage),0, -1):
                    for new_each in edges[str(k-1)]:
                        if new_each[1] == tempcluster:
                            track[k]= [new_each[1]]#[edges[str(k-1)][tempcluster][1]]
                            tempcluster = new_each[0] #edges[str(k-1)][tempcluster][0] 
                            break
                    # track[k]= [edges[str(k-1)][tempcluster][1]]
                    
                tempcluster = [clusterid]
                
                for k in range(int(stage)+1,self.total_stage):
                    
                    track[k]=self.getDescendants(tempcluster,k,edges)
                    tempcluster=track[k]
        return track

    def getZandZc(self,adata,stage,cluster,CUDA = False, impactfactor=None,topN = None):
        if stage < len(self.hiddenReps):
            if impactfactor is None:
                return self.hiddenReps[stage]
            else:
                data = self.perturb_stage_data_mean[stage]
        else:
            
            

            clusterAdataID = adata.obs[adata.obs['leiden'] == str(cluster)].index.tolist()
            clusterAdata = adata[clusterAdataID]
            
            # gcn_connectivities = clusterAdata.obsp['gcn_connectivities']
            
            # temp = gcn_connectivities[clusterAdataID]
            # temp = list(set(temp.indices))
            # # input = np.zeros(shape = (adata.X.shape[0],adata.X.shape[1]))
            # input = csr_matrix(input)
                
            # input[temp] = adata.X[temp]
            # input = gcn_connectivities @ input
            # input = input[temp].toarray()
            input = clusterAdata.obsp['gcn_connectivities'] @ clusterAdata.X
            if issparse(input):
                input = input.toarray()
            data = np.mean(input,axis=0).reshape(1,-1)
        loadModelDict = self.model_name#'./'+self.target_directory+'/model_save/'+self.model_name+'.pth'
        # vae = VAE(clusterAdata.shape[1], 256, 1024, 64,beta=1,distribution='ziln')#torch.load(loadModelDict)
        # vae = torch.load(loadModelDict)
        vae = VAE(data.shape[1], 64, 256, 0.5)
        if CUDA:
            vae.load_state_dict(torch.load(loadModelDict,map_location=torch.device('cuda:0')))
            # vae = torch.load(loadModelDict)
            vae.to('cuda:0')
        else:

            vae.load_state_dict(torch.load(loadModelDict), map_location=torch.device('cpu'))
            vae.to('cpu')
        
        vae.eval()
        recons = []
        zs = []
        zmeans = []
        zstds = []
        # data = np.mean(clusterAdata.layers['concat'].toarray(),axis=0).reshape(1,-1)#
        # print(np.mean(data))
        if impactfactor is not None:
            data = np.repeat(data.reshape(1,-1), len(impactfactor), axis=0)
            cell_loader = DataLoader(data.astype('float32'), batch_size=len(impactfactor), shuffle=False, num_workers=0)
        else:
            cell_loader = DataLoader(data.astype('float32'), batch_size=1, shuffle=False, num_workers=0)
        # cell_loader = DataLoader(clusterAdata.layers['concat'].toarray().astype('float32'), batch_size=2000, shuffle=False, num_workers=1)
        # xx_adj = np.ones(np.shape(clusterAdata.X)[0])
        
        adj = torch.tensor(1)
        if CUDA:
            adj = adj.to('cuda:0')
        for x in cell_loader:


            if impactfactor is not None:

                impactfactor = impactfactor.astype('float32')
                # x = x+impactfactor
                x = x+x*impactfactor

    
            # if on GPU put mini-batch into CUDA memory
            if CUDA:
                x = x.to('cuda:0')

            z=vae.getZ(x,adj,0,0,1,test=True)
            zs+=z[0].cpu().detach().numpy().tolist()
            zmeans+=z[1].cpu().detach().numpy().tolist()
            zstds+=z[2].cpu().detach().numpy().tolist()
        zs = np.array(zs)

        if stage >= len(self.hiddenReps):
            self.hiddenReps.append(zs)
            self.perturb_stage_data_mean.append(data)
        return zs
    class perturbationthread(threading.Thread):
        def __init__(self, outer_instance,outs, selectedstage,selectedcluster,track,bound,perturbated_gene,CUDA):
            threading.Thread.__init__(self)
            self.selectedstage = selectedstage
            self.selectedcluster = selectedcluster
            self.track = track
            self.bound = bound
            self.outs = outs
            self.perturbated_gene = perturbated_gene
            self.outer_instance = outer_instance
            self.CUDA = CUDA
        def run(self):
            self.outs[self.selectedstage]+=self.outer_instance.perturbation__auto_centroid(self.outer_instance.stageadata[self.selectedstage], self.outer_instance.stageadata, self.selectedstage, self.selectedcluster, self.track, self.bound,self.perturbated_gene,self.CUDA)

    def getTrackReadOrder(self):
        '''
        for each completed path in track (completed path = control->1->2->3, number of completed paths = number of 3 nodes), return a dictionary of orders. 
        like the path has stage3:1 is the second one to be read.
        '''
        path = self.idrem_dir#os.path.join(self.target_directory,'idremVizCluster')
        filenames = os.listdir(path) #defalut path
        tempTrack = [[] for _ in range(self.total_stage)]
        for each in filenames:
            temp = each.split('.')[0].split('-')
            for i,item in enumerate(temp):
                temp1 = item.split('n')
                tempTrack[i].append(temp1)
        dic = {}
        for i, ids in enumerate(tempTrack[-1]):
            for each in ids:
                dic[int(each)] = i
        return dic
    def perfect_perturbation__auto_centroid(self,stageadata, adata, selectedstage,selectedcluster,track,bound,perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        adatacollection = adata#[]
        
        plotadata = []
        for stage, clusters in enumerate(track):
            for clusterid, leiden in enumerate(clusters):
                if stage == selectedstage and leiden == selectedcluster:
                    flag = len(hiddenReps)
                
                temp = self.getZandZc(adatacollection[stage],stage,leiden,CUDA=CUDA)
            
                hiddenReps.append(temp)
        hiddenReps = np.array(hiddenReps)

        dijresults = []
        
        count=0
        
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
            
                temp=self.getDistance(hiddenReps[flag],hiddenReps[count])
                count+=1
            dijresults.append(temp)
            

        perturbated_stage = int(selectedstage)
  
        
        
        impactFactor = perturbated_genes
        impactFactor[impactFactor!=0] = 1
        temp = impactFactor[0].copy()
        perturbated_stage = int(selectedstage)
        selectedstage=1
        if bound > 0:
            mean_previous = np.mean(np.array(adatacollection[selectedstage-1].X.toarray()),axis=0)
            mean_current = np.mean(np.array(adatacollection[selectedstage].X.toarray()),axis=0)
            if selectedstage == 0:
                mean_previous = mean_current
            
            diff = mean_previous - mean_current
            por = diff/mean_current
            impactFactor = por * impactFactor
        else:

            mean_previous = np.mean(np.array(adatacollection[selectedstage+1].X.toarray()),axis=0)
            mean_current = np.mean(np.array(adatacollection[selectedstage].X.toarray()),axis=0)
            if selectedstage == len(adata)-1:
                mean_previous = mean_current
            diff = mean_previous - mean_current
            por = diff/mean_current
            impactFactor = por * impactFactor

        selectedtemp = self.getZandZc(None,selectedstage,track[selectedstage][0],impactfactor = impactFactor,CUDA=CUDA)
        count = 0
        fijresults = []
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
                temp = self.getDistance(selectedtemp,hiddenReps[count])
                count+=1
            fijresults.append(temp)
        delta = np.array(fijresults) - np.array(dijresults)

        gc.collect()
        
        out = []
        for i in range(impactFactor.shape[0]):
            temp = []
            temp.append(perturbated_genes[i])
            for kk in range(len(track)):
                temp.append(track[kk][0])
            for kk in range(len(track)):
                temp.append(delta[kk][i])
            out.append(temp)

        return out
    def perturbation__auto_centroid(self,stageadata, adata, selectedstage,selectedcluster,track,bound,perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        adatacollection = adata#[]
        
        plotadata = []
        for stage, clusters in enumerate(track):
            for clusterid, leiden in enumerate(clusters):
                if stage == selectedstage and leiden == selectedcluster:
                    flag = len(hiddenReps)
               
                temp = self.getZandZc(adatacollection[stage],stage,leiden,CUDA=CUDA)
            
                hiddenReps.append(temp)
        hiddenReps = np.array(hiddenReps)

        dijresults = []
        
        count=0
        
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
            
                temp=self.getDistance(hiddenReps[flag],hiddenReps[count])
                count+=1
            dijresults.append(temp)
        impactFactor = perturbated_genes
        selectedtemp = self.getZandZc(None,selectedstage,track[selectedstage][0],impactfactor = impactFactor,CUDA=CUDA)
        count = 0
        fijresults = []
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
                temp = self.getDistance(selectedtemp,hiddenReps[count])
                count+=1
            fijresults.append(temp)
        delta = np.array(fijresults) - np.array(dijresults)

        gc.collect()
        
        out = []
        for i in range(impactFactor.shape[0]):
            temp = []
            temp.append(perturbated_genes[i])
            for kk in range(len(track)):
                temp.append(track[kk][0])
            for kk in range(len(track)):
                temp.append(delta[kk][i])
            out.append(temp)

        return out
    def prepare_speed_perturbation_data(self, adata, stage, leiden,raw=None, impactfactors=None):
        if raw is None:
            stageadata = adata[stage]
            clusterAdata = self.stage_cluster[str(stage)][str(leiden)]#stageadata[self.stage_cluster_index[str(stage)][str(leiden)]]#.index.tolist()
            # clusterAdata = stageadata[clusterAdataID]
            # print(clusterAdata.obsp['gcn_connectivities'])
            # matrix1 = cupyx.scipy.sparse.csr_matrix(clusterAdata.obsp['gcn_connectivities'])#cp.asarray(clusterAdata.obsp['gcn_connectivities'])
            # matrix2 = cp.asarray(clusterAdata.X)
            # input = matrix1 @ matrix2
            # input = cp.asnumpy(input)
            input = clusterAdata.obsp['gcn_connectivities'] @ clusterAdata.X
            if issparse(input):
                input = input.toarray()

            data = np.mean(input,axis=0).reshape(1,-1)
        else:
            data = raw
        if impactfactors is not None:
            # print(np.mean(data + data * impactfactors- data))
            data =data + data * impactfactors
        
        if raw is None:
            return data.reshape(-1)
        else:
            return data
    def getZ_speedup(self, input,CUDA=False):
        zs = []
        zmeans = []
        zstds = []
        loadModelDict = self.model_name#'./'+self.target_directory+'/model_save/'+self.model_name+'.pth'
        vae = myGrphVAEPyroModule(input.shape[1], 64, 0, 0.5)
        if CUDA:
            
            vae.load_state_dict(torch.load(loadModelDict,map_location='cuda:0'))
            # vae = torch.load(loadModelDict)
            vae.to('cuda:0')
        else:
            vae.load_state_dict(torch.load(loadModelDict), map_location=torch.device('cpu'))
            vae.to('cpu')
        
        vae.eval()
        cell_loader = DataLoader(input.astype('float32'), batch_size=len(input), shuffle=False, num_workers=0)
        adj = torch.tensor(1)
        if CUDA:
            adj = adj.to('cuda:0')
        for x in cell_loader:
            # if on GPU put mini-batch into CUDA memory
            if CUDA:
                x = x.to('cuda:0')
            z=vae.getZ(x,adj,0,0,1,test=True)
            zs+=z[0].cpu().detach().numpy().tolist()
            zmeans+=z[1].cpu().detach().numpy().tolist()
            zstds+=z[2].cpu().detach().numpy().tolist()
        zs = np.array(zs)
        return zs
    def perturbation__auto_centroid_speed(self,adata, lastClusters, perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        input_data = []
        input_pertubred = []
        # adatacollection = adata#[]
        
        for i, each in enumerate(lastClusters):
            track = self.getTrack(len(self.stageadata)-1,each)
            for stage, clusters in enumerate(track):
                for clusterid, leiden in enumerate(clusters):
                    input_data.append(self.prepare_speed_perturbation_data(adata, stage, leiden))
        
        input_data = np.array(input_data)
        input_pertubred_forward = self.prepare_speed_perturbation_data(adata, stage, leiden, raw =input_data, impactfactors = perturbated_genes[0])
        input_pertubred_backward = self.prepare_speed_perturbation_data(adata, stage, leiden, raw =input_data, impactfactors = perturbated_genes[1])
        input_pertubred = np.append(input_pertubred_forward,input_pertubred_backward,axis=0)
        input_pertubred = np.array(input_pertubred)

        Z_input = self.getZ_speedup(input_data,CUDA).reshape(-1,4,64)
        Z_perturbed = self.getZ_speedup(input_pertubred,CUDA).reshape(-1,4,64)
        input_distance = []
        for i, each in enumerate(Z_input):
            for  j, each1 in enumerate(each):
                for k, each2 in enumerate(each):
                    each1 = each1.reshape(1, -1)
                    each2 = each2.reshape(1, -1)
                    input_distance.append(self.getDistance(each1,each2))
        input_distance = np.array(input_distance)
        input_distance = input_distance.reshape(-1,4,4)
        distance = []
        # print(input_distance.shape)
        for i, each in enumerate(Z_perturbed):
            for j, each1 in enumerate(Z_perturbed[i]):
                if i < len(Z_perturbed)//2:
                    count = i
                else:
                    count = i - len(Z_perturbed)//2
                for k, each2 in enumerate(Z_input[count]):
                    each1 = each1.reshape(1, -1)
                    each2 = each2.reshape(1, -1)
                    distance.append(self.getDistance(each1,each2))
        
        distance = np.array(distance)
        distance = distance.reshape(-1,4,4)
        # print(distance.shape)
        
        #delta1 is first half of delta, delta2 is second half of delta
        delta1 = distance[:len(distance)//2]- input_distance
        delta2 = distance[len(distance)//2:]- input_distance
        # print(delta1.shape)
        # print(delta2.shape)
        return delta1, delta2
        
        # out = []
        # for i in range(impactFactor.shape[0]):
        #     temp = []
        #     temp.append(perturbated_genes[i])
        #     for kk in range(len(track)):
        #         temp.append(track[kk][0])
        #     for kk in range(len(track)):
        #         temp.append(delta[kk][i])
        #     out.append(temp)

        return out
    def get_drug_genes(self,bound):
        drug_target = self.adata.uns['data_drug_overlap_genes']#this should be an attribute of adata object later on
        #drug_target = dict(np.load(drug_cell_type_target,allow_pickle=True).tolist()) #this should be an attribute of adata object later on
        drug_names = list(drug_target.keys())

        drug_target_genes = list(drug_target.values())
        temp_drug_target_genes = drug_target_genes.copy()
        perturbed_genes = []
        
        for temp in temp_drug_target_genes:
            
            out_temp = []
            for each in temp:
                each = each.split(':')
                if each[1] == '+':
                    each = each[0]+':'+str(bound)
                elif each[1] == '-':
                    each = each[0]+':'+str(1/bound)
                out_temp.append(each)
            perturbed_genes.append(out_temp)
        return drug_names, perturbed_genes
    def startAutoPerturbation_online(self,lastCluster,perturbed_genes,CUDA=False):
        '''
        Start the perturbation analysis (online version).

        parameters
        -------------------
        lastCluster: int
            The cluster id of the last cluster in the track
        perturbed_genes: list
            A list of perturbed genes
        CUDA: bool
            Whether to use GPU

        return
        -------------------
        out: dict
            A dictionary of perturbation results
        '''
        out = {}
        out['online'] = {}
        track = self.getTrack(len(self.stageadata)-1,lastCluster)

        perturbed_genes = [perturbed_genes]
        perturbated_gene_ids = []
        impactFactor = []
        for perturbated_gene in perturbed_genes:
            
            temp_perturbated_gene = perturbated_gene.copy()
            temp_bound = []
            perturbated_gene = []
            for each in temp_perturbated_gene:
                each = each.split(':')
                if len(each) > 1:
                    temp_bound.append(float(each[1]))
                    perturbated_gene.append(each[0])
                else:
                    perturbated_gene = temp_perturbated_gene
                    break

            perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
            perturbated_gene_ids.append(perturbated_gene_id)
            temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))

            for id_each, each in enumerate(perturbated_gene_id):
                temp[each] = temp_bound[id_each]-1
            
            impactFactor.append(temp)
        impactFactor = np.array(impactFactor)

  

        perturbation_results = [[] for i in range(len(track))]
   
        for i, selectedcluster in enumerate(track):
            # print(i)
            threads = []
            self.stageadata[i].obs['leiden'] = self.stageadata[i].obs['leiden'].astype('string')
            perturbation_results[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, None,impactFactor,CUDA)
                                      
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
            
            for od, each in enumerate(perturbation_results[i]):
                tempout = []
                for kk in range(self.total_stage):
                    tempout.append(each[len(each)-self.total_stage+kk])
                out['online'][str(i)] = tempout

        return out

    def startAutoPerturbation_online_speed(self,lastClusters,perturbed_genes,track_names,CUDA=False):
        '''
        Start the perturbation analysis (online version).
        
        parameters
        -------------------
        lastClusters: list
            A list of last clusters in the track
        perturbed_genes: list
            A list of perturbed genes
        track_names: list
            A list of track names
        CUDA: bool
            Whether to use GPU

        return
        -------------------
        out1: dict
            A dictionary of perturbation results
        out2: dict
            A dictionary of perturbation results

        '''
        out = {}
        out['online'] = {}
        
        temp_track = self.getTrack(len(self.stageadata)-1,lastClusters[0])
        impactFactors = []
        for each_perturbed_genes in perturbed_genes:
            perturbed_genes_temp = [each_perturbed_genes]
            perturbated_gene_ids = []
            impactFactor = []
            for perturbated_gene in perturbed_genes_temp:
                
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))

                for id_each, each in enumerate(perturbated_gene_id):
                    temp[each] = temp_bound[id_each]-1
                
                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)
            impactFactors.append(impactFactor)
  

        perturbation_results = [[] for i in range(len(temp_track))]
        outs = self.perturbation__auto_centroid_speed(self.stageadata,lastClusters,impactFactors,CUDA)
        out1 = {}
        out2 = {}
        for i, each in enumerate(lastClusters):
            track = self.getTrack(len(self.stageadata)-1,each)

            out1[track_names[i]] = {}
            out2[track_names[i]] = {}
            out1[track_names[i]]['online'] = {}
            out2[track_names[i]]['online'] = {}
            for j, selectedcluster in enumerate(track):
                out1[track_names[i]]['online'][str(j)] = outs[0][i][j]
                out2[track_names[i]]['online'][str(j)] = outs[1][i][j]
        
        return out1, out2
    
    def assign_random_direction_to_random_genes(self,random_genes):
        '''
        Build the the sets of random genes with random direction.

        parameters
        -------------------
        random_genes: list
            A of list of random genes

        return
        -------------------
        out: list
            A list of random genes with random direction
        reversed_out: list
            A list of random genes with reversed direction
        '''
        out = []
        reversed_out = []
        temp_random_genes = random_genes.copy()
        for temp in temp_random_genes:
            out_temp = []
            copyout_temp = []
            for each in temp:
                copyeach = each
                flag = np.random.choice(['-','+'])
                # bound = np.random.uniform(1.00001,3)
                bound = 3
                if flag == '+':
                    
                    each = each+':'+str(bound)
                    copyeach = copyeach+':'+str(1/bound)
                elif flag == '-':
                    each = each+':'+str(1/bound)
                    copyeach = copyeach+':'+str(bound)
                out_temp.append(each)
                copyout_temp.append(copyeach)
            out.append(out_temp)
            reversed_out.append(copyout_temp)
        return out,reversed_out
   #~~~~~         
    def startAutoPerturbation(self,lastCluster,bound,mode,CUDA = True,random_genes= None, random_times = None,written=True):
        '''
        Start the perturbation analysis.

        parameters
        -------------------
        lastCluster: int
            The cluster id of the last cluster in the track
        bound: float    
            The perturbation bound
        mode: str
            The perturbation mode, can be 'drug', 'pathway', 'perfect', 'random_background', 'online_random_background'
        CUDA: bool
            Whether to use GPU
        random_genes: list
            A list of random genes
        random_times: int
            The number of random genes
        written: bool
            Whether to write the results to disk

        return
        -------------------
        None
        '''
        track = self.getTrack(len(self.stageadata)-1,lastCluster)

        track_name = str(track[0][0]) 
        for i in range(1,len(track)):
            track_name += '-' + str(track[i][0])
        outs = [[] for i in range(len(track))]
        if mode == 'drug':
            perturbed_items, perturbed_genes = self.get_drug_genes(bound)
        elif mode == 'pathway':
            pathway_gene = self.adata.uns['data_pathway_overlap_genes']
            #pathway_gene = dict(np.load('./'+self.target_directory+'/data_pathway_overlap_genes.npy',allow_pickle=True).tolist())#this should be an attribute of adata object later on
            perturbed_items = list(pathway_gene.keys())
            temp_perturbed_genes = list(pathway_gene.values())
            perturbed_genes = []
            for each in temp_perturbed_genes:
                if type(each)!= list:
                    each = each.tolist()

                perturbed_genes.append(each)
        
        ######
        elif mode == 'perfect':
            perturbed_items, perturbed_genes = self.get_drug_genes(bound)

        elif mode == 'random_background':
        
            perturbated_genes = []
            genelen = len(self.stageadata[0].var.index.tolist())
            genenames = np.array(list(self.stageadata[0].var.index.values))
            shuffled_gene_id = [j for j in range(genelen)]
        elif mode == 'online_random_background':
            bound = 'A'
            genelen = len(self.stageadata[0].var.index.tolist())
            genenames = np.array(list(self.stageadata[0].var.index.values))
            shuffled_gene_id = [j for j in range(genelen)]
            perturbed_genes = []
            perturbed_items = []
            
            for each in range(random_times):
                random_genes = random.randint(1,3)
                perturbed_items.append(str(each))
                random.shuffle(shuffled_gene_id)
                shuffled_gene = genenames[shuffled_gene_id[:random_genes]]
                perturbed_genes.append(shuffled_gene.tolist())
            perturbed_genes,reversed_perturbed_genes = self.assign_random_direction_to_random_genes(perturbed_genes)
        if mode != 'random_background':
            impactFactor = []
            perturbated_gene_ids = []
            for perturbated_gene in perturbed_genes:

                if type(perturbated_gene) != list:
                    perturbated_gene = [perturbated_gene]
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                if len(temp_bound) == 0:
                    temp[perturbated_gene_id] = 1
                    temp = temp*(bound-1)
                else:
                    for id_each, each in enumerate(perturbated_gene_id):
                        temp[each] = temp_bound[id_each]-1

                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)


        for i, selectedcluster in enumerate(track):
    
            if '%s_perturbation_deltaD'%mode not in self.adata.uns.keys():
                self.adata.uns['%s_perturbation_deltaD'%mode] = {}
            threads = []
            if mode == 'random_background':
                perturbed_genes = []
                perturbed_items = []
                for each in range(random_times):
                    perturbed_items.append(str(each))
                    random.shuffle(shuffled_gene_id)
                    shuffled_gene = genenames[shuffled_gene_id[:random_genes]]
                    perturbed_genes.append(shuffled_gene.tolist())
                impactFactor = []
                perturbated_gene_ids = []
                for perturbated_gene in perturbed_genes:
                
                    if type(perturbated_gene) != list:
                        perturbated_gene = [perturbated_gene]
                    temp_perturbated_gene = perturbated_gene.copy()
                    temp_bound = []
                    perturbated_gene = []
                    for each in temp_perturbated_gene:
                        each = each.split(':')
                        if len(each) > 1:
                            temp_bound.append(float(each[1]))
                            perturbated_gene.append(each[0])
                        else:
                            perturbated_gene = temp_perturbated_gene
                            break

                    perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                    perturbated_gene_ids.append(perturbated_gene_id)
                    temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                    if len(temp_bound) == 0:
                        temp[perturbated_gene_id] = 1
                        temp = temp*(bound-1)
                    else:
                        for id_each, each in enumerate(perturbated_gene_id):
                            temp[each] = temp_bound[id_each]-1
                    
                    impactFactor.append(temp)
                impactFactor = np.array(impactFactor)

            self.stageadata[i].obs['leiden'] = self.stageadata[i].obs['leiden'].astype('string')
            # outs[i] = self.perturbationthread(self,i, selectedcluster[0], track, bound,impactFactor,CUDA)
            if mode == 'perfect':
                outs[i] +=self.perfect_perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)

            else:
                outs[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)

            for od, each in enumerate(outs[i]):
                if written == True:
                    # with open('./'+self.target_directory+'/tsdg2_%s_%s.csv'%(mode,bound),"a+") as f:
                        # f.write('%s,%d-%d-%d-%d,%.7f,%.7f,%.7f,%.7f\n'%(perturbed_items[od],each[1], each[2], each[3], each[4],each[5],each[6],each[7],each[8]))
                    pass #fix this later on
                else:
                    
                    # for  name in perturbed_items:
                    if str(bound) not in self.adata.uns['%s_perturbation_deltaD'%mode].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)] = {}
                    if track_name not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name] = {}
                    if perturbed_items[od] not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]] = {}
                    if str(i) not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]].keys():
                        tempout = []
                        for kk in range(self.total_stage):
                            tempout.append(each[len(each)-self.total_stage+kk])
                        # if self.mode == 'random_background':
                        #     if bound < 1/bound:
                        #         bound = 1/bound
                        #     self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]+str(bound)][str(i)] = tempout
                        # else:
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]][str(i)] = tempout
        if mode == 'online_random_background':
            bound = 'B'
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
            reversed_perturbed_genes
            impactFactor = []
            perturbated_gene_ids = []
            for perturbated_gene in reversed_perturbed_genes:
            
                if type(perturbated_gene) != list:
                    perturbated_gene = [perturbated_gene]
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                if len(temp_bound) == 0:
                    temp[perturbated_gene_id] = 1
                    temp = temp*(bound-1)
                else:
                    for id_each, each in enumerate(perturbated_gene_id):
                        temp[each] = temp_bound[id_each]-1
                
                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)
        
            outs = [[] for i in range(len(track))]
            for i, selectedcluster in enumerate(track):
                threads = []
                outs[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)
                # threads.append(self.perturbationthread(self,outs, i, selectedcluster[0], track, bound,impactFactor,CUDA))

                for od, each in enumerate(outs[i]):
                    if str(bound) not in self.adata.uns['%s_perturbation_deltaD'%mode].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)] = {}
                    if track_name not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name] = {}
                    if perturbed_items[od] not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]] = {}
                    if str(i) not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]].keys():
                        tempout = []
                        for kk in range(self.total_stage):
                            tempout.append(each[len(each)-self.total_stage+kk])
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]][str(i)] = tempout
    def run(self,mode,log2fc,inplace=False,random_times = 100,random_genes = 2,CUDA = True):
        '''
        Perform perturbation.

        parameters
        -------------------
        mode: str
            perturbation mode, 'drug', 'pathway', 'random_background', 'online_random_background', 'perfect'
        log2fc: float
            log2fc of the perturbation
        inplace: bool
            whether to write the perturbation results to the adata object
        random_times: int
            number of random genes to be perturbed
        random_genes: int
            number of random genes to be perturbed
        CUDA: bool
            whether to use CUDA

        return
        -------------------
        None
        '''
        if inplace == False:
            written=True
        else:
            written=False
        if mode == 'drug':
            for i in self.tracks.keys():

                print(i)
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'pathway':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'random_background':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,log2fc,mode,written = written,random_times=random_times,random_genes=random_genes,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode, written = written,random_times=random_times,random_genes=random_genes,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'online_random_background':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,1,mode,random_times=random_times,written =written,CUDA=CUDA)
        elif mode == 'perfect':
            for i in self.tracks.keys():

                print(i)
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
    def run_online_speed(self, allTracks:bool,perturbated_gene,perturbated_gene_reversed, unit_name,stage = None, lastCluster=None,CUDA=False):
        '''
        Perform online perturbation.
        
        parameters
        -------------------
        allTracks: bool
            Using one track or all tracks

        perturbated_gene: dict
            gene to be perturbed and the regulated intensity ({a:0.5, b: 2.5, c:0.5...})
        perturbated_gene_reversed: dict
            gene to be perturbed and the regulated intensity ({a:2.0, b: 0.4, c:2.0...} (reversed log2fc to the original)

        unit_name: str
            name of the unit to be perturbed
        stage: 
            stage to be perturbed, if None choose all
        CUDA: bool
            whether to use CUDA

        return
        -------------------
        perturbation_score: float
            perturbation score
        pval: float
            p value
        out_deltaD: dict
            deltaD of the perturbed unit
        '''
        import time

        online_analyst = perturbationAnalysis(self.adata,self.idrem_dir,stage=stage,allTracks=allTracks)
        # else:

            # online_analyst = perturbationAnalysis(self.adata,os.path.join(self.target_directory,'idremVizCluster'),allTracks=allTracks)
        perturbated_gene = perturbated_gene.split(',')
        perturbated_gene_reversed = perturbated_gene_reversed.split(',')
        if allTracks != True:
            out1 = {}
            out2 = {}
            track = self.getTrack(len(self.stageadata)-1,lastCluster)
            track_name = str(track[0][0])
            for i in range(1,len(track)):
                track_name += '-' + str(track[i][0])
            
            out1[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene,CUDA=CUDA)
            out2[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene_reversed,CUDA=CUDA)
            track = self.getTrack(len(self.stageadata)-1,lastCluster)

            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([track_name,[out1,out2]])
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
        else:
            out1 = {}
            out2 = {}
            last_clusters = []
            track_names = []
            for i in self.tracks.keys():
            
                track = self.getTrack(len(self.stageadata)-1,i)
                last_clusters.append(track[-1][0])
                track_name = str(track[0][0]) 
                for i in range(1,len(track)):
                    track_name += '-' + str(track[i][0])
                track_names.append(track_name)
            
            out1,out2 = self.startAutoPerturbation_online_speed(last_clusters,[perturbated_gene,perturbated_gene_reversed],track_names,CUDA=CUDA)
            # out2 = self.startAutoPerturbation_online_speed(last_clusters,perturbated_gene_reversed,track_names,CUDA=CUDA)
               
            step6_start = time.time()
            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([out1,out2])
            step6_end = time.time()
            print('step6 time: ', step6_end - step6_start)
        return perturbation_score, pval,out_deltaD
    def run_online(self, allTracks:bool,perturbated_gene,perturbated_gene_reversed, unit_name,stage = None, lastCluster=None,CUDA=False):
        '''
        Perform online perturbation.

        parameters
        -------------------
        allTracks: bool
            One track or all tracks
        stage: int
            stage to be perturbed
        lastCluster: int
            last cluster to be perturbed (if allTracks is False)
        perturbated_gene: list
            gene to be perturbed format a:0.5, b: 2.5, c:0.5...
        perturbated_gene_reversed: list
            gene to be perturbed format a:2.0, b: 0.4, c:2.0... (reversed log2fc to the original)
        unit_name: str
            name of the unit to be perturbed
        stage: int
            stage to be perturbed, if None choose all
        CUDA: bool
            whether to use CUDA

        return
        -------------------
        perturbation_score: np.float
            perturbation score
        pval: np.float
            p value
        '''
        import time

        online_analyst = perturbationAnalysis(self.adata,self.idrem_dir,stage=stage,allTracks=allTracks)
        # else:

            # online_analyst = perturbationAnalysis(self.adata,os.path.join(self.target_directory,'idremVizCluster'),allTracks=allTracks)
        perturbated_gene = perturbated_gene.split(',')
        perturbated_gene_reversed = perturbated_gene_reversed.split(',')
        if allTracks != True:
            out1 = {}
            out2 = {}
            track = self.getTrack(len(self.stageadata)-1,lastCluster)
            track_name = str(track[0][0])
            for i in range(1,len(track)):
                track_name += '-' + str(track[i][0])
            
            out1[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene,CUDA=CUDA)
            out2[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene_reversed,CUDA=CUDA)
            track = self.getTrack(len(self.stageadata)-1,lastCluster)

            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([track_name,[out1,out2]])
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
        else:
            out1 = {}
            out2 = {}
            for i in self.tracks.keys():
            
                track = self.getTrack(len(self.stageadata)-1,i)
                track_name = str(track[0][0])
                for j in range(1,len(track)):
                    track_name += '-' + str(track[j][0])
         
                out1= self.startAutoPerturbation_online(track[-1][0],perturbated_gene,CUDA=CUDA)
                out2 = self.startAutoPerturbation_online(track[-1][0],perturbated_gene_reversed,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
            step6_start = time.time()
            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([out1,out2])
            step6_end = time.time()
            print('step6 time: ', step6_end - step6_start)
        return perturbation_score, pval,out_deltaD
    def analysis(self,mode,log2fc,all=True,stage=None):
        '''
        Analysis of perturbation results
        
        parameters
        ----------------
        mode: str
            The mode is choosing from ['drug', 'pathway', 'online']
        log2fc: float
            log2fc is the log2 fold change of perturbation
        all: bool
            all is whether to analysis all tracks or one track
        stage: int
            stage is the stage to be analysis, if all is True, stage is None
        '''

        self.adata.obs['leiden'] = self.adata.obs['leiden'].astype('string')
        self.adata.obs['stage'] = self.adata.obs['stage'].astype('string')
        analyst = perturbationAnalysis(self.adata,self.idrem_dir, stage=stage,log2fc = log2fc, mode = mode)
       
        
        temp = analyst.main_analysis(track_to_analysis = 'all', all=all, score='avg_backScore', items=None)#read item from disk for now
        if '%s_perturbation_score'%mode not in self.adata.uns.keys():
            self.adata.uns['%s_perturbation_score'%mode] = {}
        
        self.adata.uns['%s_perturbation_score'%mode][str(log2fc)] = temp
        self.adata.uns['%s_perturbation_score'%mode][str(1/log2fc)] = temp
        


import pandas as pd
def get_top_pathways(adata, intensity, top_n=None, cutoff=None):
    """
    Get top pathways predictions after pathway perturbations at a given intensity.

    Parameters
    ----------
    adata : AnnData object
        Annotated data matrix.
    intensity : int
        Pertubration intensity.
    top_n : int, optional
        Number of top pathways to return. The default is None.
    cutoff : float, optional
        P-value cutoff. The default is None.
    """


    if top_n is not None:
        temp = pd.DataFrame.from_dict(adata.uns['pathway_perturbation_score'][str(intensity)]['total']['top_compounds'])
        temp.rename(columns={'compound': 'pathways'}, inplace=True)
        temp.rename(columns={'drug_regulation': 'regulated genes'}, inplace=True)
        return temp[:top_n]
    elif cutoff is not None:
        temp = pd.DataFrame.from_dict(adata.uns['pathway_perturbation_score'][str(intensity)]['total']['top_compounds'])
        temp.rename(columns={'compound': 'pathways'}, inplace=True)
        temp.rename(columns={'drug_regulation': 'regulated genes'}, inplace=True)
        return temp[temp['pval_adjusted'] < cutoff]
    
    else:
        print('Please specify top_n or cutoff')


'''
This is the main module of UNAGI. It contains the UNAGI class, which is the main class of UNAGI. It also contains the functions to prepare the data, start the model training and start analysing the perturbation results. Initially, `setup_data` function should be used to prepare the data. Then, `setup_training`` function should be used to setup the training parameters. Finally, `run_UNAGI` function should be used to start the model training. After the model training is done, `analyse_UNAGI` function should be used to start the perturbation analysis.
'''
import subprocess
import numpy as np
from .utils.attribute_utils import split_dataset_into_stage, get_all_adj_adata
import os
import scanpy as sc
import gc
from .utils.gcn_utils import get_gcn_exp
from .train.runner import UNAGI_runner
import torch
from .model.models import VAE,Discriminator
from .UNAGI_analyst import analyst
from .train.trainer import UNAGI_trainer
class UNAGI:
    '''
    The UNAGI class is the main class of UNAGI. It contains the function to prepare the data, start the model training and start analysing the perturbation results.
    '''
    def __init__(self,):
        self.CPO_parameters = None
        self.iDREM_parameters = None
        self.species = 'human'
    def setup_data(self, data_path,stage_key,total_stage,gcn_connectivities=False,neighbors=25,threads = 20):
        '''
        The function to specify the data directory, the attribute name of the stage information and the total number of time stages of the time-series single-cell data. If the input data is a single h5ad file, then the data will be split into multiple h5ad files based on the stage information. The function can take either the h5ad file or the directory as the input. The function will check weather the data is already splited into stages or not. If the data is already splited into stages, the data will be directly used for training. Otherwise, the data will be split into multiple h5ad files based on the stage information. The function will also calculate the cell graphs for each stage. The cell graphs will be used for the graph convolutional network (GCN) based cell graph construction.
        
        parameters
        --------------
        data_path: str 
            the directory of the h5ad file or the folder contains data.
        stage_key: str
            the attribute name of the stage information.
        total_stage: int
            the total number of time stages of the time-series single-cell data.
        gcn_connectivities: bool
            whether the cell graphs are already calculated. Default is False.
        neighbors: int
            the number of neighbors for each cell used to construct the cell neighbors graph, default is 25.
        threads: int
            the number of threads for the cell graph construction, default is 20.
        '''
        if total_stage <= 2:
            raise ValueError('The total number of stages should be larger than 2')
        
        if os.path.isfile(data_path):
            self.data_folder = os.path.dirname(data_path)
        else:
            self.data_folder = data_path
        #os.path.dirname(data_path)
        self.stage_key = stage_key
        if os.path.exists(os.path.join(self.data_folder ,'0.h5ad')):
            temp = sc.read(os.path.join(self.data_folder , '0.h5ad'))
            self.input_dim = temp.shape[1]
            if 'gcn_connectivities' not in list(temp.obsp.keys()):
                gcn_connectivities = False
            else:
                gcn_connectivities = True
        else:
            print('The dataset is not splited into stages, please use setup_data function to split the dataset into stages first')
            self.data_path = data_path
            split_dataset_into_stage(self.data_path, self.data_folder, self.stage_key)
            gcn_connectivities = False
            
        self.data_path = os.path.join(data_path,'0.h5ad')
        
        self.ns = total_stage
        #data folder is the folder that contains all the h5ad files
        self.data_folder = data_path#os.path.dirname(data_path)
        dir1 = os.path.join(self.data_folder , '0')
        dir2 = os.path.join(self.data_folder , '0/stagedata')
        dir3 = os.path.join(self.data_folder , 'model_save')
        initalcommand = 'mkdir '+ dir1 +' && mkdir '+dir2 +' && mkdir '+dir3
        p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)

        if not gcn_connectivities:
            print('Cell graphs not found, calculating cell graphs for individual stages! Using K=%d and threads=%d for cell graph construction'%(neighbors,threads))
            self.calculate_neighbor_graph(neighbors,threads)
        
        
    def calculate_neighbor_graph(self, neighbors=25,threads = 20):
        '''
        The function to calculate the cell graphs for each stage. The cell graphs will be used for the graph convolutional network (GCN) based cell graph construction.
        
        parameters
        --------------
        neighbors: int
            the number of neighbors for each cell, default is 25.
        threads: int
            the number of threads for the cell graph construction, default is 20.
        '''
        get_gcn_exp(self.data_folder, self.ns ,neighbors,threads= threads)
    def setup_training(self,
                 task, 
                 dist,
                 device=None,
                 epoch_iter=10,
                 epoch_initial=20,
                 lr=1e-4,
                 lr_dis = 5e-4,
                 beta=1,
                 hidden_dim=256,
                 latent_dim=64,
                 graph_dim=1024,
                 BATCHSIZE=512,
                 max_iter=10,
                 GPU=False):
        '''
        Set up the training parameters and the model parameters.
        
        parameters
        --------------
        task: str
            the name of this task. It is used to name the output folder.
        dist: str
            the distribution of the single-cell data. Chosen from 'ziln' (zero-inflated log normal), 'zinb' (zero-inflated negative binomial), 'zig' (zero-inflated gamma), and 'nb' (negative binomial).
        device: str
            the device to run the model. If GPU is enabled, the device should be specified. Default is None.
        epoch_iter: int
            the number of epochs for the iterative training process. Default is 10.
        epoch_initial: int
            the number of epochs for the inital iteration. Default is 20.
        lr: float
            the learning rate of the VAE model. Default is 1e-4.
        lr_dis: float
            the learning rate of the discriminator. Default is 5e-4.
        beta: float
            the beta parameter of the beta-VAE. Default is 1.
        hiddem_dim: int
            the hidden dimension of the VAE model. Default is 256.
        latent_dim: int
            the latent dimension of the VAE model. Default is 64.
        graph_dim: int
            the dimension of the GCN layer. Default is 1024.
        BATCHSIZE: int
            the batch size for the model training. Default is 512.
        max_iter: int
            the maximum number of iterations for the model training. Default is 10.
        GPU: bool
            whether to use GPU for the model training. Default is False.
        '''
        self.dist = dist
        self.device = device
        self.epoch_iter = epoch_iter
        self.epoch_initial = epoch_initial
        self.lr = lr
        self.beta = beta
        self.lr_dis = lr_dis
        self.task = task
        self.latent_dim = latent_dim
        self.graph_dim = graph_dim
        self.hidden_dim = hidden_dim
        self.BATCHSIZE = BATCHSIZE
        self.max_iter = max_iter
        self.GPU = GPU
        if self.GPU:
            assert self.device is not None, "GPU is enabled but device is not specified"
            self.device = torch.device(self.device)
        else:
            self.device = torch.device('cpu')

        self.model = VAE(self.input_dim, self.latent_dim, self.hidden_dim,beta=1,distribution=self.dist)
        # self.dis_model = Discriminator(self.input_dim)
        self.unagi_trainer = UNAGI_trainer(self.model,self.task,self.BATCHSIZE,self.epoch_initial,self.epoch_iter,self.device,self.lr, self.lr_dis,cuda=self.GPU)
    def register_CPO_parameters(self,anchor_neighbors=10, max_neighbors=30, min_neighbors=5, resolution_min=0.8, resolution_max=1.2):
        '''
        The function to register the parameters for the CPO analysis. The parameters will be used to perform the CPO analysis.
        
        parameters
        --------------
        anchor_neighbors: int
            the number of neighbors for each anchor cell.
        max_neighbors: int
            the maximum number of neighbors for each cell.
        min_neighbors: int
            the minimum number of neighbors for each cell.
        resolution_min: float
            the minimum resolution for the Leiden community detection.
        resolution_max: float
            the maximum resolution for the Leiden community detection.
        '''
        self.CPO_parameters = {}
        self.CPO_parameters['anchor_neighbors'] = anchor_neighbors
        self.CPO_parameters['max_neighbors'] = max_neighbors
        self.CPO_parameters['min_neighbors'] = min_neighbors
        self.CPO_parameters['resolution_min'] = resolution_min
        self.CPO_parameters['resolution_max'] = resolution_max
    def register_species(self,species):
        '''
        The function to register the species of the single-cell data.
        
        parameters
        --------------
        species: str
            the species of the single-cell data.
        '''
        if species not in ['human','mouse']:
            raise ValueError('species should be either human or mouse')
        self.species = species

    def register_iDREM_parameters(self,Normalize_data = 'Log_normalize_data', Minimum_Absolute_Log_Ratio_Expression = 0.5, Convergence_Likelihood = 0.001, Minimum_Standard_Deviation = 0.5):
        '''
        The function to register the parameters for the iDREM analysis. The parameters will be used to perform the iDREM analysis.
        
        parameters
        --------------
        Normalize_data: str
            the method to normalize the data. Chosen from 'Log_normalize_data' (log normalize the data), 'Normalize_data' (normalize the data), and 'No_normalize_data' (do not normalize the data).
        Minimum_Absolute_Log_Ratio_Expression: float
            the minimum absolute log ratio expression for the iDREM analysis.
        Convergence_Likelihood: float
            the convergence likelihood for the iDREM analysis.
        Minimum_Standard_Deviation: float
            the minimum standard deviation for the iDREM analysis.
        '''
        
        self.iDREM_parameters = {}
        if Normalize_data not in ['Log_normalize_data','Normalize_data','No_normalize_data']:
            raise ValueError('Normalize_data should be chosen from Log_normalize_data, Normalize_data and No_normalize_data')
        self.iDREM_parameters['Normalize_data'] = Normalize_data
        self.iDREM_parameters['Minimum_Absolute_Log_Ratio_Expression'] = Minimum_Absolute_Log_Ratio_Expression
        self.iDREM_parameters['Convergence_Likelihood'] = Convergence_Likelihood
        self.iDREM_parameters['Minimum_Standard_Deviation'] = Minimum_Standard_Deviation

    def run_UNAGI(self,idrem_dir):
        '''
        The function to launch the model training. The model will be trained iteratively. The number of iterations is specified by the `max_iter` parameter in the `setup_training` function.
        
        parameters
        --------------
        idrem_dir: str
            the directory to the iDREM tool which is used to reconstruct the temporal dynamics.
        '''
        for iteration in range(0,self.max_iter):
            
            if iteration != 0:
                dir1 = os.path.join(self.data_folder , str(iteration))
                dir2 = os.path.join(self.data_folder , str(iteration)+'/stagedata')
                dir3 = os.path.join(self.data_folder , 'model_save')
                initalcommand = 'mkdir '+ dir1 +' && mkdir '+dir2
                p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
            unagi_runner = UNAGI_runner(self.data_folder,self.ns,iteration,self.unagi_trainer,idrem_dir)
            unagi_runner.set_up_species(self.species)
            if self.CPO_parameters is not None:
                if type (self.CPO_parameters) != dict:
                    raise ValueError('CPO_parameters should be a dictionary')
                else:
                    unagi_runner.set_up_CPO(anchor_neighbors=self.CPO_parameters['anchor_neighbors'], max_neighbors=self.CPO_parameters['max_neighbors'], min_neighbors=self.CPO_parameters['min_neighbors'], resolution_min=self.CPO_parameters['resolution_min'], resolution_max=self.CPO_parameters['resolution_max'])
            if self.iDREM_parameters is not None:
                if type (self.iDREM_parameters) != dict:
                    raise ValueError('iDREM_parameters should be a dictionary')
                else:
                    unagi_runner.set_up_iDREM(Minimum_Absolute_Log_Ratio_Expression = self.iDREM_parameters['Minimum_Absolute_Log_Ratio_Expression'], Convergence_Likelihood = self.iDREM_parameters['Convergence_Likelihood'], Minimum_Standard_Deviation = self.iDREM_parameters['Minimum_Standard_Deviation'])
            unagi_runner.run()

  
    def analyse_UNAGI(self,data_path,iteration,progressionmarker_background_sampling_times,target_dir=None,customized_drug=None,cmap_dir=None):
        '''
        Perform downstream tasks including dynamic markers discoveries, hierarchical markers discoveries, pathway perturbations and compound perturbations.
        
        parameters
        ---------------
        data_path: str
            the directory of the data (h5ad format, e.g. org_dataset.h5ad).
        iteration: int
            the iteration used for analysis.
        progressionmarker_background_sampling_times: int
            the number of times to sample the background cells for dynamic markers discoveries.
        target_dir: str
            the directory to save the results. Default is None.
        customized_drug: str
            the customized drug perturbation list. Default is None.
        cmap_dir: str
            the directory to the cmap database. Default is None.
        '''
        analysts = analyst(data_path,iteration,target_dir=target_dir,customized_drug=customized_drug,cmap_dir=cmap_dir)
        analysts.start_analyse(progressionmarker_background_sampling_times)
        print('The analysis has been done, please check the outputs!')


        

        


        




'''
This module contains functions to identify dynamic genes and accumulate gene weights.
'''
import numpy as np
import gc
import os
import json
import pandas as pd
import scanpy as sc
from scipy.sparse import lil_matrix
from scipy.sparse import csr_matrix
import gc

def getLastStageIDs(json,total_stage):
    '''
    get the list of node ids in last stage 
    parameters
    -----------
    json: idrem meta
    total_stage: total number of stages

    return
    -----------
    out: a list of node ids in last stage
    '''
    out = []
    for each in json[0][1:]:
        if each['nodetime'] == 'stage'+str(total_stage-1):
            out.append(each['nodeID'])
    return out
def getidremPath(json,cid):
    '''
    given the last node id in idrem, get the whole paths
    
    parameters
    -----------
    json: idrem meta file
    cid: the last node id in idrem

    return
    -----------
    nodes: a list of node ids in the path
    '''
    nodes = []

    for each in reversed(json[0][1:]):

        if each['nodeID'] == cid:
            nodes.append(cid)
            if each['parent']!= -1:
                cid = each['parent']
            else:
                break
    nodes.reverse()
    return nodes
def getIdremPaths(json,total_stage):
    '''
    get all paths in idrem
    
    parameters
    -----------
    json: list
        idrem meta files

    return
    -----------
    paths: a list of all paths
    '''
    
    stage3IDs = getLastStageIDs(json,total_stage) 
    paths = []
    for i in stage3IDs:
        paths.append(getidremPath(json, i))
    return paths
def getPosNegMaxChangesNodes(json,paths,total_stage):
    means = []
    for path in paths:
        pathmean = []
        for each in path:
            
            for node in json[0][1:]:
                if node['nodeID'] == each:
                    pathmean.append(node['nodeMean'])
        means.append(pathmean)
    means = np.array(means)
    indicator = total_stage-1
    while indicator > 0:
        means[:,indicator] = means[:,indicator]-means[:,indicator-1]
        indicator-=1
    poschangenodesid = np.argmax(means,axis=0)
    negchangenodesid = np.argmax(-means,axis=0)
    poschangenodes = []
    negchangenodes = []
    for i, each in enumerate(poschangenodesid):
        poschangenodes.append(paths[each][i])
    for i, each in enumerate(negchangenodesid):
        negchangenodes.append(paths[each][i])
    return poschangenodes, negchangenodes
def testChanges(path,filename):
    '''
    a test function
    '''
    tt = readIdremJson(path, filename)
    paths = getIdremPaths(tt,4)

    return getPosNegMaxChangesNodes(tt,paths)
def getTargetGenes(path,N):
    '''
    get top N genes of each path
    parameters
    -----------
    path: the file path of IDREM results
    
    return
    -----------
    out: a list of top N up or down regulators of each path
    '''
    out = []

    filenames = os.listdir(path)
    
    for each in filenames:
        if each[0] != '.':
            #out.append(getMaxMinPathGenes(path,each,N)) #get the genes from nodes that have highest and lowest nodemean
            out.append(getPosNegDynamicPathGenes(path,each,N)) #get the genes from nodes increase and descrease most between stages
    return out

def getPosNegDynamicPathGenes(path,filename,topN):
    '''
    get the genes from nodes that increase and decrease most between stages
    
    parameters
    -----------
    path: the file path of IDREM results
    filename: the file name of IDREM results
    topN: the number of genes to accumulate gene weights

    return
    -----------
    out: a list of top N up or down regulators of each path
    '''
    total_stage = len(filename.split('.')[0].split('-'))
    tt = readIdremJson(path, filename)
    paths = getIdremPaths(tt,total_stage)
    posdynamicids, negdynamicids = getPosNegMaxChangesNodes(tt,paths,total_stage-1)
    posdynamicgenes,posdynamicgeneids = getMaxOrMinNodesGenes(tt,posdynamicids)
    negdynamicgenes,negdynamicgeneids = getMaxOrMinNodesGenes(tt,negdynamicids)
    posnegdynamicgenes = [posdynamicgenes[i]+negdynamicgenes[i] for i in range(total_stage-1)]
    posnegdynamicgeneids = [posdynamicgeneids[i]+negdynamicgeneids[i] for i in range(total_stage-1)]
    out = getTopNTargetGenes(tt,posnegdynamicgenes,posnegdynamicgeneids,topN, total_stage-1)
    return out

def getMaxOrMinNodesGenes(json,nodes):
    '''
    get the genes from dynamic nodes
    parameters
    -----------
    json: idrem meta
    nodes: the list of dynamic nodes

    return
    -----------
    genes: a list of genes in the nodes
    '''
    boolgenes = []
    genes = []
    geneids = []

    for each in json[0]:
        if each['nodeID'] in nodes:
            boolgenes.append(each['genesInNode'])
    for each in boolgenes:
        tempgenes = []
        tempgeneids = []
        for i, gene in enumerate(each):
            if gene == True:
                tempgenes.append(json[3][i].upper())
                tempgeneids.append(i)
        genes.append(tempgenes)
        geneids.append(tempgeneids)
    return genes,geneids

def readIdremJson(path, filename):
    '''
    Parse the IDREM json file
    parameters
    -----------
    path: the file path of IDREM results
    filename: the file name of IDREM results

    return
    -----------
    tt: the parsed IDREM json file
    '''
    print('getting Target genes from ', filename)
    path = os.path.join(path,filename,'DREM.json')
    f=open(path,"r")
    lf=f.readlines()
    f.close()
    lf="".join(lf)
    lf=lf[5:-2]+']'
    tt=json.loads(lf,strict=False)
    return tt

def getTopNTargetGenes(json,genenames,geneids,topN,total_stage):
    '''
    get top N genes of each path sorted by the change of gene expression between stages

    parameters
    -----------
    json: the parsed IDREM json file
    genenames: a list of genes in the nodes
    geneids: a list of gene ids in the nodes
    topN: the number of genes to accumulate gene weights
    total_stage: the total number of stages

    return
    -----------
    out: a list of top N up or down regulators of each path
    '''

    out = [[] for i in range(total_stage)]
    for i in range(total_stage):
        changegene = np.array([json[5][j] for j in geneids[i]])
        change = abs(changegene[:,i+1]-changegene[:,i])
        pddata = pd.DataFrame(change, columns=['change_Value'])
        pddata.index = genenames[i]
        sortedchange = pddata.sort_values(by = 'change_Value',ascending=False)
        topNGenes = sortedchange.index.tolist()[:topN]
        out[i] = topNGenes
    return out

def listTracks(mid, iteration,total_stage):
    '''
    list all tracks in the selected iteration
    parameters
    -----------
    mid: directory to the task
    iteration: the selected iteration
    total_stage: the total number of stages

    return
    -----------
    tempTrack: a list of tracks
    '''
    filenames= os.listdir(os.path.join(mid,str(iteration)+'/idremResults/'))
    #filenames = os.listdir('./reresult/idremVizCluster0.1-nov14/') #defalut path
    tempTrack = [[] for _ in range(total_stage)]
    for each in filenames:
        temp = each.split('.')[0].split('-')
        for i,item in enumerate(temp):
            temp1 = item.split('n')
            tempTrack[i].append(temp1)
    return tempTrack

def matchTFandTGWithFoldChange(TFs,scopes,avgCluster,filename,genenames,total_stage):
    '''
    use target genes from IDREM as scopes to count tf and tgs with human-encode
    parameters
    -----------
    TFs: list
        A list of top N up or down regulators of each path
    scopes: list
        A list of target genes of each path
    avgCluster: list
        A list of average gene expression of each cluster
    filename: str
        The name of IDREM results
    genenames: list
        A list of gene names
    total_stage: int
        The total number of stages

    return
    -----------
    TFTG: a list of gene weights of each path
    '''
    f = open(filename,'r')
    rl = f.readlines()
    f.close()
    TG = {}
    for line in rl[1:]:
        item = line.split('\t')
        if item[0] not in TG.keys():
            TG[item[0]] = []
        genename = item[1].split(';')[0]
        #if genename in genenames:
        TG[item[0]].append(genename)
    genedict = {}
    for i, each in enumerate(genenames):
        genedict[each.upper()] = i
    TFTG = [[] for _ in range(len(TFs))]
    for i, track in enumerate(TFs):
        temp = [[] for _ in range(total_stage-1)]
        TFTG[i] = temp
        
        for j, stage in enumerate(track):
            TFTG[i][j] = np.zeros(shape=len(genenames))
            for tf in stage:
                
                targetGenes = TG[tf[0].split(' ')[0]]
                for each in scopes[i][j]:
                    foldChange = abs(np.log2(avgCluster[j+1][j][genedict[each]]+1) - np.log2(avgCluster[j][i][genedict[each]]+1))
                    TFTG[i][j][genedict[each]] = 1*foldChange+1
                temp2 = np.zeros(shape=len(genenames))
                for each in enumerate(targetGenes):
                    
                    
                    if each[1] in genedict.keys() and each[1] in scopes[i][j]:
                    
                    

                        foldChange = abs(np.log2(avgCluster[j+1][j][genedict[each[1]]]+1) - np.log2(avgCluster[j][i][genedict[each[1]]]+1))
                        # if TFTG[i][j][genedict[each[1]]]<2:
                        TFTG[i][j][genedict[each[1]]] = 1*foldChange+1 #hyperparameter needed to be adusted by users
                    elif each[1] in genedict.keys() and each[1] not in scopes[i][j]:
                        
                        foldChange = abs(np.log2(avgCluster[j+1][i][genedict[each[1]]]+1)-np.log2(avgCluster[j][i][genedict[each[1]]]+1))
                        
                        # if TFTG[i][j][genedict[each[1]]]<2:
                        TFTG[i][j][genedict[each[1]]] = 0.5*foldChange+0.5 #hyperparameter needed to be adusted by users

                if tf[0].split(' ')[0] in genenames:
                    index = genedict[tf[0].split(' ')[0]]
                    # if avgCluster[j+1][i][index] > avgCluster[j][i][index]:
                    #     if avgCluster[j][i][index] == 0:
                    #         foldChange = 1
                    #     else:
                    #         foldChange = avgCluster[j+1][i][index]/avgCluster[j][i][index]
                    # elif avgCluster[j+1][i][index] < avgCluster[j][i][index]:
                    #     if avgCluster[j+1][i][index] == 0:
                    #         foldChange = 1
                    #     else:
                    foldChange = abs(np.log2(avgCluster[j][i][index]+1)-np.log2(avgCluster[j+1][i][index]+1))
                    TFTG[i][j][index] = 2*foldChange+2
                #TFTG[i][j].append(temp2)
            #TFTG[i][j] = np.array(TFTG[i][j])
            #TFTG[i][j] = np.sum(TFTG[i][j],axis = 0)
    TFTG = np.array(TFTG)
    return TFTG

def updateGeneFactorsWithDecay(adata, clusterid,iteration,geneWeight, decayRate = 0.5):
    '''
    update gene weights and decay the weight of genes that are not important in this iteration of a cluster

    parameters
    ----------------------
    adata: anndata
        the cluster of single cell data
    clusterid: int
        the cluster id
    iteration: int
        the selected iteration
    geneWeight: np.array
        the gene weights
    decayRate: float
        the decay rate

    return
    ----------------------
    adata: anndata
        the updated single-cell data
    '''
    geneWeight = geneWeight.reshape(1,-1)
    adata.obs['leiden']=adata.obs['leiden'].astype('int64')
    cells = adata.obs.reset_index()
    celllist = cells[cells['leiden']==int(clusterid)].index.tolist()
    if 'geneWeight' not in adata.layers.keys():
        adata.layers['geneWeight'] = lil_matrix(np.zeros(adata.X.shape)) 
    else:
        adata.layers['geneWeight'] = adata.layers['geneWeight'].tolil()
    adata.layers['geneWeight'][celllist] = adata.layers['geneWeight'][celllist]*decayRate#.multiply(decayCandidate) change it to all decrease for all cells in one stage
    # adata.layers['geneWeight'][celllist] += geneWeight
    adata.layers['geneWeight'] = adata.layers['geneWeight'].tocsr()
    return adata
    
def updataGeneTablesWithDecay(mid, iteration, geneFactors,total_stage, decayRate = 0.5):
    '''
    update gene weights and decay the weight of genes that are not important in this iteration

    parameters
    ----------------------
    mid: str
        The task name
    iteration: int
        the selected iteration
    geneFactors: np.array
        the gene weights of each cell
    total_stage: int
        the total number of stages
    decayRate: float
        the decay rate

    return
    ----------------------
    difference: np.float
        the average difference of gene weights between stages
    '''
    tracks = listTracks(mid,iteration,total_stage)
    difference = 0
    for i, stage in enumerate(tracks):
        if i != 0:
            gc.collect()
            adata = sc.read_h5ad(os.path.join(mid,str(iteration)+'/stagedata/%d.h5ad'%(i)))
            temppreviousMySigmoidGeneWeight = adata.layers['geneWeight']
            adata.layers['geneWeight'] = adata.layers['geneWeight'].multiply(decayRate) #sep8 change to all decrease for all cell
            #adata = sc.read_h5ad('./stagedata/%d.h5ad'%(i))
            for j, clusterids in enumerate(stage):#in some tracks, the one stage have many clusters
                for clusterid in clusterids:
                    # print('number of idrem file', j)
                    # print('stage',i)
                    adata = updateGeneFactorsWithDecay(adata,clusterid,iteration,geneFactors[j][i-1])
                    
                    # difference+=clusterDiff#each update will get the mySigmoid(increment) 
            previousMySigmoidGeneWeight = mySigmoid(temppreviousMySigmoidGeneWeight.toarray())
            currentMySigmoidGeneWeight = mySigmoid(adata.layers['geneWeight'].toarray())
            difference += np.mean(np.absolute(currentMySigmoidGeneWeight - previousMySigmoidGeneWeight))
            adata.layers['geneWeight'] = adata.layers['geneWeight'].tocsr()
            adata.write(os.path.join(mid,str(iteration)+'/stagedata/%d.h5ad'%i),compression='gzip' )
        else:
            if int(iteration) == 0:
                adata = sc.read_h5ad(os.path.join(mid,str(iteration)+'/stagedata/%d.h5ad'%(i)))
                if 'geneWeight' not in adata.layers.keys():
                    adata.layers['geneWeight'] = csr_matrix(np.zeros(adata.X.shape)) 
                    adata.write(os.path.join(mid,str(iteration)+'/stagedata/%d.h5ad'%i),compression='gzip' )
    return difference/(total_stage-1)
def checkupDown(idrem, genename):
    '''
    check if the TFs is a up or down regulator
    parameters
    -----------
    idrem: list
        the iDREM metafile
    genename: list
        a list of name of TF(str)
    
    return
    -----------
    flag: bool
        up regulator return 1,  down regulator return 0
    '''
    genename=genename.split(' ')[0]

    if genename not in idrem[3]:
        flag = 0
    else:
        flag = 1
    return flag

def mergeTFs(TFs,total_stage):
    '''
    merge top N up or down regulators into the stage level and remove the repeated regulators among sibling nodes of IDREM tree
    parameters 
    ----------------------
    TFs: list
        a list of top N up or down regulators of a IDREM tree
    total_stage: int
         the total number of stages

    return
    -----------
    out: list
         a list of up or down regulators of each stage
    '''
    upAndDownset= [set(),set(),set()]
    out = [[] for _ in range(total_stage-1)]
    for i, each in enumerate(TFs):
        for item in each:
            for data in item:
                if data[0].split(' ')[0] not in upAndDownset[i]:
                    upAndDownset[i].add(data[0].split(' ')[0])
                    out[i].append(data)
    return out
    
def getTopNUpandDown(TFs,topN=20):
    '''
    obtain top 20 up or down regulators based on the score overall (P value)
    parameters 
    -----------
    TFs: list
        a list of up or down regulators
    topN: int
        the number of top regulators to be extracted. Default is 20
    
    return
    ----------- 
    TFs: list
        top N up or down regulators
    '''
   
    if len(TFs[0]) == 6:
        TFs = sorted(TFs,key=lambda x:x[5])
    else:
        
        TFs = sorted(TFs,key=lambda x:x[6])
    TFs = TFs[:topN]
    return TFs
def extractTFs(path,filename,total_stage,topN=20):
    '''
    extract top N up or down TFs of a certain path from the DREM json file
    parameters 
    -----------
    filename: str
        the name of certain paths
    total_stage: int
        the total number of stages
    topN: int
        the number of top regulators to be extracted. Default is 20
    
    return
    -----------
    extractedTFs: list
        top N up or down TFs of a certain path
    '''
    print('getting TFs from ', filename)
    path = os.path.join(path,filename,'DREM.json')

    extractedTFs = [[] for _ in range(total_stage-1)]
    f=open(path,"r")
    lf=f.readlines()
    f.close()
    lf="".join(lf)
    lf=lf[5:-2]+']'
    tt=json.loads(lf,strict=False)
    
    TFs = [[] for _ in range(total_stage-1)]
    stages = [str(i+1) for i in range(total_stage-1)]
    for each in tt[0][1:]:
        temp = []
        for item in each['ETF']:
            if checkupDown(tt, item[0]):
                
                temp.append(item)
        if len(temp) ==0:
            continue
        temp = getTopNUpandDown(temp,topN)
        print(each['nodetime'],stages.index(each['nodetime']))
        TFs[stages.index(each['nodetime'])].append(temp)

    extractedTFs = mergeTFs(TFs,total_stage=total_stage)
    
    return extractedTFs
def getTFs(path,total_stage,topN=20):
    '''
    get top N up or down regulators of each path
    parameters
    -----------
    path: str
        the file path of IDREM results
    topN: int
        the number of top regulators to be extracted. Default is 20
    total_stage: int
        the total number of stages
    
    return
    -----------
    out: list
        a list of top N up or down regulators of each path
    '''
    out = []
 
    filenames = os.listdir(path)
    
    for each in filenames:
        if each[0] != '.':
            out.append(extractTFs(path,each,topN=topN,total_stage=total_stage))
    return out

def mySigmoid(z,weight=-4):
    '''
    new shifted sigmoid transformation for replace strategy
    parameters 
    -----------
    z: np.array
        input data
    weight: float
        the weight of sigmoid transformation

    return
    ----------- 
    out: np.array
        data after shifted sigmoid transformation
    '''
    out = 1/(1+20*np.exp(weight*z+1.5)) #hyper parameters needed to be adjusted by users
    return out
if __name__ == '__main__':
    getPosNegDynamicPathGenes('/mnt/md0/yumin/to_upload/UNAGI/tutorials/example_1/idrem','4-4-5-1.txt_viz',50)


import scanpy as sc
import os
import numpy as np
import torch
# from pyro.contrib.examples.util import MNIST
import torch.nn as nn
import torchvision.transforms as transforms
import scanpy as sc
import torch.nn.functional as F
import pyro
import pyro.distributions as dist
# import pyro.contrib.examples.util  # patches torchvision
from pyro.infer import SVI, Trace_ELBO
from pyro.optim import Adam
import matplotlib.pyplot as plt
from torchvision.utils import make_grid
from torch.utils.data import Dataset,DataLoader
from scipy.sparse import csc_matrix
import umap
from ..dynamic_regulatory_networks.processTFs import mySigmoid
from scvi import data
import scvi
#NUMGENES=len(adata.var)
# customized h5ad dataset
class H5ADataSet(Dataset):
    '''
    The customized dataset for the data without gene weights. (For the initial iteration) The dataset will return the gene expression and the cell graph for each cell.
    '''
    def __init__(self,fname):
        self.data=fname
        # self.neighbors = neighbors
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):

        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        return x_tensor,self.data.obsp['gcn_connectivities'][idx].indices.tolist(),idx#,self.data.obs['Sample.ID'][idx]
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data

class H5ADataSetGeneWeight(Dataset):
    '''
    The customized dataset for the data with gene weights. The dataset will return the gene expression and the gene weight for each cell.
    '''
    def __init__(self,fname):
        self.data=fname
    
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):
        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        gw = self.data.layers['geneWeight'][idx][0].toarray()[0]
        gw = mySigmoid(gw.astype(np.float32))
        gw_tensor = torch.from_numpy(gw)
        return x_tensor, gw_tensor
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data




# colodict = {}
# for each in zip(all_types,category_colors):
#     colodict[each[0]] = rgb2hex(each[1])
import scanpy as sc
import gc

import matplotlib.pyplot as plt
from sklearn.metrics import silhouette_score
from sklearn.neighbors import kneighbors_graph
from sklearn.metrics import adjusted_rand_score,normalized_mutual_info_score
from sklearn.neighbors import kneighbors_graph

def plot_with_colormap(values,color_dict):
    '''
    The color scheme the cell types are plotted with.
    
    Parameters
    ----------
    values : list
        List of cell types.
    color_dict : dict
        Dictionary of cell types and their colors.
    
    Returns
    -------
    color_dict : dict
        Dictionary of cell types and their colors.
    '''
    color_list = [[0.36862745, 0.30980392, 0.63529412, 1.        ],'tab:pink','tab:olive','tab:cyan','gold', 'springgreen','coral','skyblue','tab:blue','tab:orange','tab:green','tab:red','tab:purple','tab:brown','yellow','aqua', 'turquoise','orangered', 'lightblue','darkorchid', 'fuchsia','royalblue','slategray', 'silver', 'teal', 'fuchsia','grey','indigo','khaki','magenta','tab:gray']
    # random.shuffle(color_list)
    values = list(set(values))
    values = sorted(values)
    for i, value in enumerate(values):
        if value not in list(color_dict.keys()):
            color_dict[value] = color_list[(len(list(color_dict.keys()))+1)]
    return color_dict
def plot_stages_latent_representation(adatas, cell_type_key, stage_key,color_scheme=None,ax=None,dpi=300,save=None):
    '''
    Plot the latent representation of the cells colored by cell type and leiden clusters.

    Parameters
    ----------
    adatas : AnnData object
        Annotated data matrix.
    cell_type_key : str
        Key for cell type column in adata.obs.
    stage_key : str
        Key for stage column in adata.obs.
    color_scheme : dict, optional
        Dictionary of cell types and their colors. The default is None.
    ax : matplotlib axis, optional  
        The default is None.
    dpi : int, optional
        The default is 300.
    save : str, optional
        Path to save the figure. The default is None.

    Returns
    --------------

    '''
    sc.set_figure_params(scanpy=True, dpi=dpi)
    consistency = []
    ariss= []
    NMIs = []
    silhouettes = []
    # ITERATION= 5
    stage_keys = adatas.obs[stage_key].unique().tolist()
    stage_keys = sorted(stage_keys)
    stage_keys = stage_keys[::-1]

    if color_scheme is None:
        color_dict_unagi = {}
    else:
        color_dict_unagi = color_scheme

    color_dict_leiden = {}
    color_dict_groundtruth = {}
    total_adata = 0
    count=0
    NMI = 0
    silhouettes =0
    aris = 0

    fig, ax = plt.subplots(4,2, figsize=(10,15))
    for i,stage in enumerate(stage_keys):
        
        temp_count = 0
        print(len(adatas.obs[adatas.obs[stage_key] == str(stage)].index.tolist()))
        adata = adatas[adatas.obs[adatas.obs[stage_key] == str(stage)].index.tolist()]
    #         print(len(adata))
        adata.obs['UNAGI'] = adata.obs[cell_type_key].astype('category')
        
        # adata.obs['Ground Truth'] = adata.obs['name.simple'].astype('category')
    
        adata.obs['leiden'] = adata.obs['leiden'].astype('category')

        sorted_list = sorted(list(adata.obs['UNAGI'].unique()))
        color_dict_unagi = plot_with_colormap(sorted_list,color_dict_unagi)
        adata.obs['leiden'] = adata.obs['leiden'].astype('string')
        
        sc.pl.umap(adata,color='UNAGI',ax=ax[i,0], show=False,palette=color_dict_unagi,title=stage_keys[i])
        sc.pl.umap(adata,color='leiden',ax=ax[i,1], show=False,title = stage_keys[i])
        total_adata+=len(adata)
        count+=temp_count
        print('ARI: ', adjusted_rand_score(adata.obs['name.simple'],adata.obs['UNAGI'] ))
        print('NMIs: ', normalized_mutual_info_score(adata.obs['name.simple'],adata.obs['UNAGI']))
        print('silhouette score: ', silhouette_score(adata.obsm['z'], adata.obs['leiden']))
        NMI += normalized_mutual_info_score(adata.obs['name.simple'], adata.obs['UNAGI'])
        silhouettes+=silhouette_score(adata.obsm['z'], adata.obs['leiden'])
        aris += adjusted_rand_score( adata.obs['name.simple'],adata.obs['UNAGI'])
    consistency.append(count/total_adata)
    ariss.append(aris/4)
    NMIs.append(NMI/4)
    plt.tight_layout()
    if save is not None:
        plt.savefig(save,dpi=dpi)
    else:
        plt.show()
    print('ARIs: ', ariss)
    print('NMI: ', NMIs)
    print('silhouette score: ', silhouettes/4)
if __name__ == '__main__':
    adata = sc.read_h5ad('small_1/dataset.h5ad')
    plot_stages_latent_representation(adata,'ident','stage')


import scanpy as sc


def hierarchical_static_markers_heatmap(adata,stage,cluster,level,n_genes,stage_key,celltype_key,protein_encoding_gene=True,min_logfoldchange=1,dpi=None,save=None):
    '''
    Plot the heatmap of the hierarchical static markers of the chosen cluster and its siblings.

    Parameters
    ----------
    adata : AnnData object
        Annotated data matrix.
    stage : int
        Time point of the data.
    cluster : int
        Cluster id.
    level : int
        Level of the cluster.
    n_genes : int
        Number of genes to plot.
    stage_key : str
        Key for stage column in adata.obs.
    celltype_key : str
        Key for cell type column in adata.obs.
    protein_encoding_gene : bool, optional
        Whether to plot only protein encoding genes. The default is True.
    min_logfoldchange : int, optional
        Minimum log fold change. The default is 1.
    dpi : int, optional
        The default is None.
    save : str, optional
        Path to save the figure. The default is None.

    Returns
    ----------------
    None
    '''
    import warnings
    warnings.filterwarnings('ignore')
    if protein_encoding_gene:
        protein_coding_genes = []
        for each in adata.var.index:
            if each[:2] != 'AC' and each[:2] != 'AL' and each[:2] != 'AP' and each[:4] != 'LINC' and '.' not in each:
                protein_coding_genes.append(each)
        adata = adata[:,protein_coding_genes]
    if dpi is None:
        sc.set_figure_params(scanpy=True, dpi=80,dpi_save=300)
    else:
        sc.set_figure_params(scanpy=True, dpi=80,dpi_save=dpi)
    adata.obs[stage_key] = adata.obs[stage_key].astype(str)
    adata = adata[adata.obs[stage_key] == str(stage)]
    temp_clustertype = {}
    choice = adata.uns['hcmarkers'][str(stage)]['markers'][str(cluster)][str(level)]['ref'][0]
    scope = adata.uns['hcmarkers'][str(stage)]['markers'][str(cluster)][str(level)]['ref'][1]
    scope.remove(choice[0])
    adata.obs['leiden'] = adata.obs['leiden'].astype(int)
    choice_adata = adata[adata.obs['leiden'].isin(choice)]
    scope_adata = adata[adata.obs['leiden'].isin(scope)]
    choice_adata.obs['subset'] = 'chosen'
    scope_adata.obs['subset'] = 'siblings'
    apat = choice_adata.concatenate(scope_adata)
    apat.obs['subset'] = apat.obs['subset'].astype('category')
    apat.layers['scaled'] = sc.pp.scale(apat, copy=True).X
    marker_genes = []
    pos = []
    for i in range(len(temp_clustertype.keys())):
        temp = [i*10,i*10+9]
        pos.append(tuple(temp))
    sc.tl.rank_genes_groups(apat,'subset', method='wilcoxon',  n_genes=n_genes)
    apat.obs['leiden'] = apat.obs['leiden'].astype(str)
    marker_genes = apat.uns['rank_genes_groups']['names']['chosen']
    if save is not None:
        sc.pl.heatmap(apat, marker_genes, groupby=['subset',celltype_key,'leiden'],layer='scaled',swap_axes=True,vmin=-0.5, vmax=0.5, cmap='RdBu_r',figsize=(20, 16),save = save)
    else:
        # sc.pl.heatmap(apat, marker_genes, groupby=['subset',celltype_key,'leiden'],layer='scaled',swap_axes=True,vmin=-0.5, vmax=0.5, cmap='RdBu_r',figsize=(20, 16))
        sc.pl.rank_genes_groups_heatmap(apat,groupby=['subset',celltype_key,'leiden'],layer='scaled', min_logfoldchange=min_logfoldchange,n_genes=n_genes,swap_axes=True, use_raw=False,show_gene_labels=True,vmin=-0.5, vmax=0.5, cmap='RdBu_r',var_group_positions=pos,dendrogram=False,var_group_rotation=0,figsize=(20, 16))

if __name__ == '__main__':
    import pickle
    adata = sc.read_h5ad('small_1/dataset.h5ad')
    adata.uns = pickle.load(open('small_1/attribute.pkl', 'rb'))
    hierarchical_static_markers_heatmap(adata ,stage=0,choice=3,level=1,n_genes=20, stage_key = 'stage', celltype_key='ident', protein_encoding_gene=True, min_logfoldchange=1,dpi=100, save='test.png')


from typing import Optional, Sequence, Dict, Any
import scipy.cluster.hierarchy as sch
import scanpy as sc
import numpy as np
from pandas.api.types import CategoricalDtype
import pandas as pd
from scanpy.tools._utils import _choose_representation
from anndata import AnnData
from ..dynamic_graphs.distDistance import fitClusterGaussianRepresentation,calculateKL

def getclusterreps(data):
    '''
    Get the sampled Norm distribution of Z space for each cluster

    Parameters
    ----------
    data: pandas dataframe
        The hidden representation of Z space. The index of the dataframe is the cluster id. The columns of the dataframe are the hidden representation of Z space.

    Returns
    -------
    outs: dictionary
        The sampled Norm distribution of Z space. The keys of the dictionary are the cluster id. The values of the dictionary are the sampled Norm distribution of Z space.

    '''
    clusterid = data.index.unique().tolist()
    out = {}
    for each in clusterid:
        out[each] = fitClusterGaussianRepresentation(data.loc[each].T.values)
    return out
def calculateDistance(reps):
    '''
    Calculate the distance between the sampled Norm distribution of Z space

    Parameters
    ----------
    reps: dictionary
        The sampled Norm distribution of Z space. The keys of the dictionary are the cluster id. The values of the dictionary are the sampled Norm distribution of Z space.

    Returns
    -------
    df: pandas dataframe
        The distance between the sampled Norm distribution of Z space. The index of the dataframe is the cluster id. The columns of the dataframe are the cluster id.
    '''
    outs = {}
    for i, each1 in enumerate(list(reps.keys())):
        outs[each1] = []
        for j, each2 in enumerate(list(reps.keys())):
            outs[each1].append(((calculateKL(reps[each1], reps[each2])+calculateKL(reps[each2], reps[each1]))/2))
    df = pd.DataFrame(outs,index=list(reps.keys()))
    df.columns = list(reps.keys())
    df.index = df.index.astype(int)
    df = df.sort_index()
    df.columns =  df.columns.astype(int)
    df = df.sort_index(axis=1)
    return df
def calculateDistanceUMAP(umaps):
    '''
    Calculate the distance between umap coordinates of Z space

    Parameters
    ----------
    umaps: pandas dataframe
        The umap coordinates of Z space. The index of the dataframe is the cluster id. The columns of the dataframe are the umap coordinates of Z space.

    Returns 
    -------
    df: pandas dataframe
        The distance between umap coordinates of Z space. The index of the dataframe is the cluster id. The columns of the dataframe are the cluster id.
    '''
    outs = {}

    for i, each1 in enumerate(list(umaps.index)):
        outs[each1] = []
        
        for j, each2 in enumerate(list(umaps.index)):
            outs[each1].append(np.linalg.norm(umaps.loc[each1]- umaps.loc[each2]))
    df = pd.DataFrame(outs,index=list(umaps.index))
    df.columns = list(umaps.index)
    df.index = df.index.astype(int)
    df = df.sort_index()
    df.columns =  df.columns.astype(int)
    df = df.sort_index(axis=1)
    return df
def mydendrogram(
    adata: AnnData,
    groupby: str,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    var_names: Optional[Sequence[str]] = None,
    use_raw: Optional[bool] = None,
    cor_method: str = "pearson",
    linkage_method: str = "complete",
    optimal_ordering: bool = False,
    key_added: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, Any]]:
    """\
    Modified from scanpy.tl.dendrogram. 
    Computes a hierarchical clustering for the given `groupby` categories.

    By default, the latent representation 'Z' is used.

    Alternatively, a list of `var_names` (e.g. genes) can be given.

    Average values of either `var_names` or components are used
    to compute a correlation matrix.

    The hierarchical clustering can be visualized using
    :func:`scanpy.pl.dendrogram` or multiple other visualizations that can
    include a dendrogram: :func:`~scanpy.pl.matrixplot`,
    :func:`~scanpy.pl.heatmap`, :func:`~scanpy.pl.dotplot`,
    and :func:`~scanpy.pl.stacked_violin`.

    .. note::
        The computation of the hierarchical clustering is based on predefined
        groups and not per cell. The correlation matrix is computed using by
        default pearson but other methods are available.

    Parameters
    ----------
    adata
        Annotated data matrix
    {n_pcs}
    {use_rep}
    var_names
        List of var_names to use for computing the hierarchical clustering.
        If `var_names` is given, then `use_rep` and `n_pcs` is ignored.
    use_raw
        Only when `var_names` is not None.
        Use `raw` attribute of `adata` if present.
    cor_method
        correlation method to use.
        Options are 'pearson', 'kendall', and 'spearman'
    linkage_method
        linkage method to use. See :func:`scipy.cluster.hierarchy.linkage`
        for more information.
    optimal_ordering
        Same as the optimal_ordering argument of :func:`scipy.cluster.hierarchy.linkage`
        which reorders the linkage matrix so that the distance between successive
        leaves is minimal.
    key_added
        By default, the dendrogram information is added to
        `.uns[f'dendrogram_{{groupby}}']`.
        Notice that the `groupby` information is added to the dendrogram.
    inplace
        If `True`, adds dendrogram information to `adata.uns[key_added]`,
        else this function returns the information.

    Returns
    -------
    If `inplace=False`, returns dendrogram information,
    else `adata.uns[key_added]` is updated with it.
    """
    if isinstance(groupby, str):
        # if not a list, turn into a list
        groupby = [groupby]
    for group in groupby:
        if group not in adata.obs_keys():
            raise ValueError(
                "groupby has to be a valid observation. "
                f"Given value: {group}, valid observations: {adata.obs_keys()}"
            )
        if not isinstance(adata.obs[group].dtype, CategoricalDtype):
            raise ValueError(
                "groupby has to be a categorical observation. "
                f"Given value: {group}, Column type: {adata.obs[group].dtype}"
            )

    if var_names is None:
        rep_df = pd.DataFrame(
            _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs)
        )
        categorical = adata.obs[groupby[0]]
        if len(groupby) > 1:
            for group in groupby[1:]:
                # create new category by merging the given groupby categories
                categorical = (
                    categorical.astype(str) + "_" + adata.obs[group].astype(str)
                ).astype("category")
        categorical.name = "_".join(groupby)

        rep_df.set_index(categorical, inplace=True)
        categories = rep_df.index.categories
    else:
        gene_names = adata.raw.var_names if use_raw else adata.var_names

    # aggregate values within categories using 'mean'
    
    mean_df = rep_df.groupby(level=0).mean()

    import scipy.cluster.hierarchy as sch
    from scipy.spatial import distance
    test = getclusterreps(rep_df)
    test2 = calculateDistance(test)
    corr_matrix = mean_df.T.corr(method=cor_method)
    corr_condensed = distance.squareform(test2)

    z_var = sch.linkage(
        corr_condensed, method=linkage_method, optimal_ordering=optimal_ordering
    )
    dendro_info = sch.dendrogram(z_var, labels=list(categories), no_plot=True)

    dat = dict(
        linkage=z_var,
        groupby=groupby,
        use_rep=use_rep,
        cor_method=cor_method,
        linkage_method=linkage_method,
        categories_ordered=dendro_info["ivl"],
        categories_idx_ordered=dendro_info["leaves"],
        dendrogram_info=dendro_info,
        correlation_matrix=corr_matrix.values,
    )

    if inplace:
        if key_added is None:
            key_added = f'dendrogram_{"_".join(groupby)}'
        # logg.info(f"Storing dendrogram info using `.uns[{key_added!r}]`")
        adata.uns[key_added] = dat
    else:
        return dat
def mydendrogramUMAP(
    adata: AnnData,
    groupby: str,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    var_names: Optional[Sequence[str]] = None,
    use_raw: Optional[bool] = None,
    cor_method: str = "pearson",
    linkage_method: str = "complete",
    optimal_ordering: bool = False,
    key_added: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, Any]]:
    """\
    Modify from scanpy.tl.dendrogram but the distance is between umap coordinates of latent representation 'Z' space
    
    Computes a hierarchical clustering for the given `groupby` categories.


    Alternatively, a list of `var_names` (e.g. genes) can be given.

    Average values of either `var_names` or components are used
    to compute a correlation matrix.

    The hierarchical clustering can be visualized using
    :func:`scanpy.pl.dendrogram` or multiple other visualizations that can
    include a dendrogram: :func:`~scanpy.pl.matrixplot`,
    :func:`~scanpy.pl.heatmap`, :func:`~scanpy.pl.dotplot`,
    and :func:`~scanpy.pl.stacked_violin`.

    .. note::
        The computation of the hierarchical clustering is based on predefined
        groups and not per cell. The correlation matrix is computed using by
        default pearson but other methods are available.

    Parameters
    ----------
    adata
        Annotated data matrix
    {n_pcs}
    {use_rep}
    var_names
        List of var_names to use for computing the hierarchical clustering.
        If `var_names` is given, then `use_rep` and `n_pcs` is ignored.
    use_raw
        Only when `var_names` is not None.
        Use `raw` attribute of `adata` if present.
    cor_method
        correlation method to use.
        Options are 'pearson', 'kendall', and 'spearman'
    linkage_method
        linkage method to use. See :func:`scipy.cluster.hierarchy.linkage`
        for more information.
    optimal_ordering
        Same as the optimal_ordering argument of :func:`scipy.cluster.hierarchy.linkage`
        which reorders the linkage matrix so that the distance between successive
        leaves is minimal.
    key_added
        By default, the dendrogram information is added to
        `.uns[f'dendrogram_{{groupby}}']`.
        Notice that the `groupby` information is added to the dendrogram.
    inplace
        If `True`, adds dendrogram information to `adata.uns[key_added]`,
        else this function returns the information.

    Returns
    -------
    If `inplace=False`, returns dendrogram information,
    else `adata.uns[key_added]` is updated with it.
    """
    if isinstance(groupby, str):
        # if not a list, turn into a list
        groupby = [groupby]
    for group in groupby:
        if group not in adata.obs_keys():
            raise ValueError(
                "groupby has to be a valid observation. "
                f"Given value: {group}, valid observations: {adata.obs_keys()}"
            )
        if not isinstance(adata.obs[group].dtype, CategoricalDtype):
            raise ValueError(
                "groupby has to be a categorical observation. "
                f"Given value: {group}, Column type: {adata.obs[group].dtype}"
            )

    if var_names is None:
        rep_df = pd.DataFrame(
            _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs)
        )
        categorical = adata.obs[groupby[0]]
        if len(groupby) > 1:
            for group in groupby[1:]:
                # create new category by merging the given groupby categories
                categorical = (
                    categorical.astype(str) + "_" + adata.obs[group].astype(str)
                ).astype("category")
        categorical.name = "_".join(groupby)

        rep_df.set_index(categorical, inplace=True)
        categories = rep_df.index.categories
    else:
        gene_names = adata.raw.var_names if use_raw else adata.var_names

    # aggregate values within categories using 'mean'
    
    mean_df = rep_df.groupby(level=0).mean()

    import scipy.cluster.hierarchy as sch
    from scipy.spatial import distance
    # test = getclusterrepsUMAP(rep_df)

    test2 = calculateDistanceUMAP(mean_df)
    # corr_matrix = mean_df.T.corr(method=cor_method)
    corr_condensed = distance.squareform(test2)

    z_var = sch.linkage(
        corr_condensed, method=linkage_method, optimal_ordering=optimal_ordering
    )
    dendro_info = sch.dendrogram(z_var, labels=list(categories), no_plot=True)

    dat = dict(
        linkage=z_var,
        groupby=groupby,
        use_rep=use_rep,
        cor_method=cor_method,
        linkage_method=linkage_method,
        categories_ordered=dendro_info["ivl"],
        categories_idx_ordered=dendro_info["leaves"],
        dendrogram_info=dendro_info,
        # correlation_matrix=corr_matrix.values,
    )

    if inplace:
        if key_added is None:
            key_added = f'dendrogram_{"_".join(groupby)}'
        # logg.info(f"Storing dendrogram info using `.uns[{key_added!r}]`")
        adata.uns[key_added] = dat
    else:
        return dat
def get_siblings_at_each_level(leaf_id, Z):
    """
    Get the sibling clusters for the given leaf at each level of the hierarchical clustering.

    Parameters
    ----------  
    leaf_id: int
        The leaf id of the given leaf.
    Z: numpy.ndarray
        The linkage matrix of the hierarchical clustering.

    Returns
    -------
    siblings: list
        The sibling clusters for the given leaf at each level of the hierarchical clustering.
    """
    n = Z.shape[0] + 1  # The total number of initial leaves
    clusters = {i: [i] for i in range(n)}  # Initialize clusters with single leaves
    siblings = []

    # Helper function to find the parent node for the given leaf or cluster
    def find_parent(node):
        for i, (left, right, _, _) in enumerate(Z):
            if node in [left, right]:
                return i + n, left if node != left else right
        return None, None

    # Build clusters and collect siblings at different levels
    for k in range(n - 1):
        left, right = int(Z[k, 0]), int(Z[k, 1])
        clusters[n + k] = clusters[left] + clusters[right]

    # Trace back from the leaf to the root and collect siblings
    current_node = leaf_id
    while True:
        parent, sibling_node = find_parent(current_node)
        if parent is None:  # Reached the root
            break
        # Check if sibling_node is a leaf or a cluster and collect all leaves
        sibling_leaves = clusters[sibling_node]
        siblings.append(sibling_leaves)
        current_node = parent

    return siblings[::-1]  # Reverse to have siblings from bottom to top
def map_leaves_to_label(categories_idx_ordered, categories_ordered):
    table = {}
    for i, each in enumerate(categories_idx_ordered):
        table[each] = int(categories_ordered[i])
    return table
def my_logfold_change(adata1, adata2,log_fold_change_cutoff=None,abs = False):
    '''
    Define the log fold change between two datasets. (data is already log transformed) LogFoldChange = Log(mean2) - Log(mean1)

    Parameters
    ----------
    adata1: AnnData
        The annotated data matrix.
    adata2: AnnData
        The annotated data matrix.
    log_fold_change_cutoff: float   
        The cutoff of the log fold change to select the genes. Default is None.
    abs: bool
        If True, the absolute value of the log fold change will be used. Default is False.

    Returns
    -------
    df: pandas dataframe
        The log fold change between two datasets. The columns of the dataframe are the log fold change, the names of the genes, the mean of the genes in the first dataset and the mean of the genes in the second dataset.

    '''
    genenames = adata1.var_names.tolist()
    mean1 = np.mean(adata1.X.toarray(),axis=0)
    mean2 = np.mean(adata2.X.toarray(),axis=0)
    logfold_change = mean2 - mean1
    logfold_change_dict = {}
    order = np.argsort(logfold_change)[::-1]
    mean2 = mean2[order]
    mean1 = mean1[order]
    for i, each in enumerate(genenames):
        logfold_change_dict[each] = logfold_change[i]
    if abs ==True:
        temp = {k: v for k, v in sorted(logfold_change_dict.items(), key=lambda item: np.abs(item[1]),reverse=True)}
    else:
        temp = {k: v for k, v in sorted(logfold_change_dict.items(), key=lambda item: item[1],reverse=True)}

    df = pd.DataFrame()
    df['log_fold_changes'] = temp.values()
    df['names'] = temp.keys()

    if log_fold_change_cutoff is not None:
        df = df[np.abs(df['log_fold_changes'])>log_fold_change_cutoff]
    df['Mean_compared_clusters'] = mean1
    df['Mean_selected_clusters'] = mean2
    return df
def build_reference_siblings(dendrogram):
    '''
    Find the sibling clusters for each cluster at each level of the hierarchical clustering.

    Parameters
    ----------
    dendrogram: dict
        The data structure to store the hierarchical clustering information. The data structure is the output of the function mydendrogram.
    
    Returns
    -------
    out: dict
        The data structure to store the sibling clusters for each cluster at each level of the hierarchical clustering. The keys of the dictionary are the cluster ids. The values of the dictionary are the sibling clusters at each level of the hierarchical clustering. 
    '''
    total_leaves = len(dendrogram['dendrogram_info']['ivl'])
    table = map_leaves_to_label(dendrogram['categories_idx_ordered'], dendrogram['categories_ordered'])
    out = {}
    for leave_id in range(total_leaves):
        siblings = get_siblings_at_each_level(leave_id, dendrogram['linkage'])
        leaf_node_id = table[leave_id]
        out[str(leaf_node_id)] = {}
        level_table = {}
        memory = []
        for i, sibs in enumerate(siblings[::-1]):
            translated_sibs = [table[each] for each in sibs]
            memory+=translated_sibs
            level_table[len(siblings)-1-i] = memory.copy()
        for level, sibs in enumerate(siblings):
            out[str(leaf_node_id)][str(level)] = level_table[level]
    return out
def gethcmarkers(adata, selected,groups,cutoff=0.05):
    '''
    Get the hierarchical clustering markers for the selected cluster and the background clusters.

    Parameters
    ----------
    adata: AnnData
        The annotated data matrix.
    selected: int
        The cluster id of the selected cluster.
    groups: list    
        The cluster ids of the background clusters.
    cutoff: float
        The cutoff of the adjusted p-value to select the hierarchical clustering markers.

    Returns 
    -------
    positive_marker: pandas dataframe
        The positive hierarchical clustering markers for the selected cluster. 
    negative_marker: pandas dataframe
        The negative hierarchical clustering markers for the selected cluster.
    '''
    ref = [[selected],[selected]+groups]
    ids = []
    ids = adata[adata.obs['leiden'] == str(selected)].obs.index.tolist()
    selectedids = ids.copy()
    groupids = []
    for each in groups:
        groupids +=adata[adata.obs['leiden'] == str(each)].obs.index.tolist()
    ids += groupids.copy()
    new = adata[ids]
    new.obs['compare'] = None
    new.obs.loc[new.obs.index.isin(selectedids), "compare"] = 'selected'
    new.obs.loc[new.obs.index.isin(groupids), "compare"] = 'background'
    sc.tl.rank_genes_groups(new,groupby='compare',method='wilcoxon',rankby_abs=False)
    selected = new[selectedids]
    background = new[groupids]
    df1 = my_logfold_change(background,selected)
    df1 = df1.set_index('names')
    df = pd.DataFrame.from_dict({key: new.uns['rank_genes_groups'][key]['selected'] for key in['names','pvals_adj','scores']})
    df = df.set_index('names')
    df = df.join(df1)
    positive_marker = df[df['log_fold_changes']>0]
    negative_marker = df[df['log_fold_changes']>0]
    positive_marker= positive_marker[positive_marker['pvals_adj'] <cutoff]
    negative_marker= negative_marker[negative_marker['pvals_adj'] <cutoff]
    
    return positive_marker, negative_marker,ref
def get_stage_hcmarkers(adata,key,use_rep,cutoff=0.05):
    '''
    Perform hierarchical clustering on the stage and get the hierarchical clustering markers for each cluster.

    Parameters
    ----------
    adata: AnnData
        The annotated data matrix.
    key: str
        The key of the cluster information in adata.obs.
    use_rep: str
        The key of the representation in adata.obsm.
    cutoff: float
        The cutoff of the adjusted p-value to select the hierarchical clustering markers.

    Returns
    -------
    Z: numpy.ndarray
        The linkage matrix of the hierarchical clustering.
    order: list
        The order of the clusters in the hierarchical clustering.
    out_table: dict
        The data structure to store the hierarchical clustering markers for each cluster.
    '''
    if use_rep == 'z':
        mydendrogram(adata,use_rep='z',groupby=key)
    elif use_rep == 'umaps':
        mydendrogramUMAP(adata,use_rep='X_umap',groupby=key)
    
    test = build_reference_siblings(adata.uns['dendrogram_'+key])
    Z = adata.uns['dendrogram_'+key]['linkage']
    order = adata.uns['dendrogram_'+key]['categories_ordered']
    out_table = {}
    for each in list(test.keys()):
        
        out_table[each] = {}
        for level in test[each]:
            out_table[each][level] = {}
            out_table[each][level]['chosen'] = {}
            pos_markers, neg_markers, ref = gethcmarkers(adata,int(each),test[each][level],cutoff=cutoff)
            out_table[each][level]['chosen']['pos'] = pos_markers
            out_table[each][level]['chosen']['neg'] = neg_markers
            out_table[each][level]['ref'] = ref
    return Z,order,out_table
def get_dataset_hcmarkers(adata,stage_key,cluster_key,use_rep,cutoff=0.05):
    '''
    Perform hierarchical clustering on the dataset and get the hierarchical clustering markers for each stage.

    Parameters
    ----------
    adata: AnnData
        The annotated data matrix.
    stage_key: str
        The key of the stage information in adata.obs.
    cluster_key: str
        The key of the cluster information in adata.obs.
    use_rep: str
        The key of the representation in adata.obsm.
    cutoff: float
        The cutoff of the adjusted p-value to select the hierarchical clustering markers. Default is 0.05.

    Returns 
    -------
    hcmarkers: dict
        The hierarchical clustering markers for each stage. The keys of the dictionary are the cluster information in adata.obs. The values of the dictionary are the positive and negative hierarchical clustering markers for each cluster. The positive and negative hierarchical clustering markers are pandas dataframe with the columns of the names of the genes, the log fold change and the adjusted p-value.
    '''
    hcmarkers = {}
    for each in list(adata.obs[stage_key].unique()):
        hcmarkers[str(each)] = {}
        stage_adata = adata[adata.obs[stage_key] == each]
        Z,order,markers = get_stage_hcmarkers(stage_adata,cluster_key,use_rep=use_rep,cutoff=cutoff)
        hcmarkers[str(each)]['Z'] = Z
        hcmarkers[str(each)]['order'] = order
        hcmarkers[str(each)]['markers'] = markers
    return hcmarkers
# get_dataset_hcmarkers(adata,'stage','leiden')
#unit test
if __name__ == '__main__':
    import pickle
    # adata = sc.read_h5ad('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/dataset.h5ad')
    adata = sc.read('/mnt/md0/yumin/UNAGI_pyro/src/small_0/dataset.h5ad')
    # adata.uns = pickle.load(open('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/attribute.pkl', 'rb'))
    adata.uns = pickle.load(open('/mnt/md0/yumin/UNAGI_pyro/src/small_0/attribute.pkl', 'rb'))
    
    adata.uns['hcmarkers'] = get_dataset_hcmarkers(adata,stage_key='stage',cluster_key='leiden',use_rep='umaps')
    # with open('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/attribute_1109_newhc.pkl','wb') as f:
    #     pickle.dump(adata.uns,f)


import pandas as pd
def get_top_compounds(adata, intensity, top_n=None, cutoff=None):
    '''
    Get top compounds predictions after compound perturbations at a given intensity.

    Parameters
    ----------
    adata : AnnData object
        Annotated data matrix.
    intensity : int
        Pertubration intensity.
    top_n : int, optional
        Number of top compounds to return. The default is None.
    cutoff : float, optional
        P-value cutoff. The default is None.
    '''
    if top_n is not None:
        return pd.DataFrame.from_dict(adata.uns['drug_perturbation_score'][str(intensity)]['total']['top_compounds'])[:top_n]
    elif cutoff is not None:
        temp = pd.DataFrame.from_dict(adata.uns['drug_perturbation_score'][str(intensity)]['total']['top_compounds'])
        return temp[temp['pval_adjusted'] < cutoff]
    else:
        print('Please specify top_n or cutoff')


from setuptools import setup, find_packages

setup(
    name='scUNAGI',
    version='0.2.3',
    
    long_description=open('README.md').read(),
    packages=find_packages(),
    package_data={'UNAGI': ['data/*.npy','data/*.txt']},
    long_description_content_type='text/markdown',
    install_requires=[
    'h5py>=3.9.0',
    'idna>=3.4',
    'igraph>=0.10.7',
    'igraph<=0.11',
    'networkx>=3.1',
    'numba>=0.57.1',
    'numpy>=1.24.4',
    'pandas>=2.1.0',
    'parso>=0.8.3',
    'pillow>=9.4.0',
    'pip>=23.2.1',
    'psutil>=5.9.5',
    'ptyprocess>=0.7.0',
    'pure_eval>=0.2.2',
    'pycparser>=2.21',
    'pygments>=2.16.1',
    'pynndescent>=0.5.10',
    'pyopenssl>=23.2.0',
    'pyparsing>=3.1.1',
    'pysocks>=1.7.1',

    'python-dateutil>=2.8.2',
    'python-igraph>=0.10.8',
    'pytz>=2023.3.post1',
    'pyzmq>=25.1.1',
    'lightning==2.0.8',
    'numpyro==0.13.0',
    'requests>=2.31.0',
    'scanpy>=1.9.5',
    'scikit-learn>=1.3.0',
    'scipy>=1.9.3',
    'seaborn>=0.12.2',
    'setuptools>=68.0.0',
    'six>=1.16.0',
    'stack_data>=0.6.2',
    'statsmodels>=0.14.0',
    'stdlib-list>=0.8.0',
    'sympy>=1.11.1',
    'tbb>=2021.8.0',
    'texttable>=1.6.7',
    'threadpoolctl>=3.2.0',
    'torchaudio>=2.0.0',
    'torchvision>=0.15.0',
    'tornado>=6.3.3',
    'tqdm>=4.66.1',
    'traitlets>=5.9.0',
    'typing_extensions>=4.7.1',
    'tzdata>=2023c',
    'umap-learn>=0.5.3',
    'wcwidth>=0.2.6',
    'wheel>=0.38.4',
    'pyro-ppl>=1.8.6',
    'scvi-tools>=1.0.3',
    'leidenalg>=0.10.1',
    'recommonmark'
    ],
    include_package_data=True,
    author='Yumin Zheng',
    author_email='yumin.zheng@mail.mcgill.ca',
    description='UNAGI: Deep Generative Model for Deciphering Cellular Dynamics and In-Silico Drug Discovery in Complex Diseases',
    url='https://github.com/mcgilldinglab/UNAGI',
)
